# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields.foreign - Foreign field, links to another object
#


# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

import sys
import weakref

from . import IRField, irNull

from .null import IR_NULL_STR

from ..compat_str import isStringy, to_unicode, isBaseStringy


__all__ = ( 
	'ForeignLinkDataBase', 'ForeignLinkData', 'ForeignLinkMultiData',
	'IRForeignLinkFieldBase', 'IRForeignLinkField', 'IRForeignMultiLinkField'
)

class ForeignLinkDataBase(object):
	'''
		ForeignLinkDataBase - Base class for data relating to foreign links
	'''
	pass

class ForeignLinkData(ForeignLinkDataBase):
	'''
		ForeignLinkData - Link data for storing information about a foreign object (id and maybe object itself).

		Can fetch object if not already fetched
	'''

	__slots__ = ('pk', 'obj', '_foreignModel')

	def __init__(self, pk=None, foreignModel=None, obj=None):
		'''
			__init__ - Create a ForeignLinkData object

			@param pk <int> - The primary key of the foreign object
			@param obj <None/IndexedRedisModel> - The resolved object, or None if not yet resolved
		'''
		self.pk = pk
		self.obj = obj

		if foreignModel is not None:
			# Shouldn't share a weakref...
			if issubclass(foreignModel.__class__, weakref.ReferenceType):
				foreignModel = foreignModel()

			self._foreignModel = weakref.ref(foreignModel)
		else:
			self._foreignModel = None


					
	
	def __getitem__(self, name):
		if not isStringy(name):
			raise ValueError('Indexes must be string. %s is not supported' %( hasattr(name, '__name__') and name.__name__ or name.__class__.__name__, ) )

		if name == 'pk':
			return self.getPk()
		elif name == 'obj':
			return self.getObj()

		raise KeyError('No such item: "%s". Possible keys are:  [ "pk", "obj" ]' %(name, ) )

	@property
	def foreignModel(self):
		'''
			foreignModel - Resolve and return the weakref to the associated foreign model
		'''
		return self._foreignModel()

	def getObj(self):
		'''
			getObj - Fetch (if not fetched) and return the obj associated with this data.
		'''
		if self.obj is None:
			if not self.pk:
				return None
			self.obj = self.foreignModel.objects.get(self.pk)

		return self.obj
	
	def getObjs(self):
		'''
			getObjs - Fetch (if not fetched) and return the obj associated with this data.

				Output is iterable.
		'''
		return [ self.getObj() ]
	
	def getPk(self):
		'''
			getPk - Resolve any absent pk's off the obj's (like if an obj has been saved), and return the pk.
		'''
		if not self.pk and self.obj:
			if self.obj._id:
				self.pk = self.obj._id

		return self.pk
	
	def getPks(self):
		return [ self.getPk() ]

	def isFetched(self):
		'''
			isFetched - Check if the associated obj has been fetched or not.
		'''
		return not bool(self.obj is None)


	def __repr__(self):
		foreignModelName = self._foreignModel and self._foreignModel().__name__ or 'None'

		return self.__class__.__name__ + '(pk=%s , foreignModel=%s , obj=%s)' %(self.pk, foreignModelName, self.obj)

	def __eq__(self, other):
		if self.__class__ != other.__class__:
			return False
		
		if other.getPk() == self.getPk():
			return True

		return False
		
	def __ne__(self, other):
		return not self.__eq__(other)

	def objHasUnsavedChanges(self):
		'''
			objHasUnsavedChanges - Check if any object has unsaved changes, cascading.
		'''
		if not self.obj:
			return False

		return self.obj.hasUnsavedChanges(cascadeObjects=True)
			


# TODO: Maybe create a base which both of these extend,
#   As having multiple in a singular field name can get confusing
class ForeignLinkMultiData(ForeignLinkData):
	'''
		ForeignLinkMultiData - Link data with multiple links.

		@see ForeignLinkData
	'''

	def __init__(self, pk=None, foreignModel=None, obj=None):
		'''
			__init__ - Create a ForeignLinkMultiData

			@see ForeignLinkData
		'''
		ForeignLinkData.__init__(self, pk, foreignModel, obj)
		
		pk = self.pk
		obj = self.obj
		if pk and not obj:
			self.obj = [None for i in range(len(pk))]

		elif obj and not pk:
			self.pk = []
			for thisObj in obj:
				if thisObj._id:
					self.pk.append(thisObj._id)
				else:
					raise ValueError('Unset id')
		elif len(obj) != len(pk):
			if len(pk) > len(obj):
				self.obj += [None for i in range( len(pk) - len(obj) ) ]
			else:
				for thisObj in obj[len(pk):]:
					if thisObj._id:
						self.pk.append(thisObj._id)
					else:
						raise ValueError('unset id')


	def getPk(self):
		'''
			getPk - @see ForeignLinkData.getPk
		'''
		if not self.pk or None in self.pk:
			for i in range( len(self.pk) ):
				if self.pk[i]:
					continue

				if self.obj[i] and self.obj[i]._id:
					self.pk[i] = self.obj[i]._id

		return self.pk

	def getPks(self):
		return self.getPk()


	def getObj(self):
		'''
			getObj - @see ForeignLinkData.getObj

				Except this always returns a list
		'''
		if self.obj:
			needPks = [ (i, self.pk[i]) for i in range(len(self.obj)) if self.obj[i] is None]

			if not needPks:
				return self.obj

			fetched = list(self.foreignModel.objects.getMultiple([needPk[1] for needPk in needPks]))
			
			i = 0
			for objIdx, pk in needPks:
				self.obj[objIdx] = fetched[i]
				i += 1

		return self.obj

	def getObjs(self):
		'''
			getObjs - @see ForeignLinkData.getObjs
		'''
		return self.getObj()

	def isFetched(self):
		'''
			isFetched - @see ForeignLinkData.isFetched
		'''
		if not self.obj:
			return False

		if not self.pk or None in self.obj:
			return False
		return not bool(self.obj is None)


	def objHasUnsavedChanges(self):
		'''
			objHasUnsavedChanges - @see ForeignLinkData.objHasUnsavedChanges

			True if ANY object has unsaved changes.
		'''
		if not self.obj:
			return False

		for thisObj in self.obj:
			if not thisObj:
				continue
			if thisObj.hasUnsavedChanges(cascadeObjects=True):
				return True

		return False


class IRForeignLinkFieldBase(IRField):
	'''
		IRForeignLinkFieldBase - Base class for Foreign Link fields
	'''
	pass

class IRForeignLinkField(IRForeignLinkFieldBase):
	'''
		IRForeignLinkField - A field which provides a one-to-one mapping to another IndexedRedisModel object.

	'''

	CAN_INDEX = True

	def __init__(self, name='', foreignModel=None):
		'''
			__init__ - Create an IRForeignLinkField. Only takes a name

			@param name <str> - Field name

			This field type does not support indexing.
		'''
		IRField.__init__(self, name, valueType=int, defaultValue=irNull)

		if foreignModel:
			if not isinstance(foreignModel, type):
				raise ValueError('foreignModel must point to a type (class), not an instance.')
			if not hasattr(foreignModel, '_is_ir_model'):
				raise ValueError('foreignModel must extend IndexedRedisModel')

		self._foreignModel = weakref.ref(foreignModel)

	@property
	def foreignModel(self):
		'''
			foreignModel - Resolve and return the weakref to the associated Foreign Model
		'''
		return self._foreignModel()


	def _fromStorage(self, value):
		value = super(IRForeignLinkField, self)._fromStorage(value)

		if not value:
			return value

		return ForeignLinkData(value, self.foreignModel)

	def _fromInput(self, value):

		if hasattr(value, '_is_ir_model'):
			return ForeignLinkData(value._id, self.foreignModel, value)

		elif isinstance(value, int):

			return ForeignLinkData(int(value), self.foreignModel)

		elif isBaseStringy(value) and value.isdigit():

			return ForeignLinkData(int(value), self.foreignModel)

		elif value in (None, irNull):

			return irNull

		elif issubclass(value.__class__, ForeignLinkData):

			return value

		raise ValueError('Unknown input (expected either irNull, int (pk), or %s. Got: <%s>   %s' %(self.foreignModel.__name__, value.__class__.__name__, repr(value)) )
	
	def _toStorage(self, value):

		if isinstance(value, int):
			return str(value)

		elif isinstance(value, ForeignLinkData):
			pk = value.getPk()
			if pk:
				return str(pk)
			else:
				return IR_NULL_STR

		elif hasattr(value, '_is_ir_model'):
			if value._id:
				return str(value._id)
			else:
				# Linked to an unsaved object, and cascadeSave must be False... hope they know what they're doing! :)
				return IR_NULL_STR


		raise ValueError('Unknown value type headed for storage:   <%s>   %s' %(value.__class__.__name__, repr(value)))


	def _toIndex(self, value):
		# Support passing either an integer or the model itself
		if issubclass(value.__class__, IRForeignLinkField):
			return value.getPk()

		return super(IRForeignLinkField, self)._toIndex(value)

	def _getReprProperties(self):
		return [
			'valueType=%s' %(str(self.valueType), ),
			'foreignModel=%s' %(self.foreignModel.__name__, )
		]

	def copy(self):
		return self.__class__(name=self.name, valueType=self.valueType, foreignModel=self.foreignModel)

	def isMulti(self):
		'''
			isMulti - Returns True if this is a MultiLink object (expects lists), otherwise False (expects object)

			@return <bool>
		'''
		return False

	def __new__(self, name='', foreignModel=None):
		return IRField.__new__(self, name)


class IRForeignMultiLinkField(IRForeignLinkField):
	'''
		IRForeignMultiLinkField - A field which links to a list of foreign objects
	'''

	# So... technically we CAN index this. But I feel like most legit scenarios would need a "link contains" operation,
	#  while direct-redis filtering only supports equals and not equals.. so really they're looking at client-side filtering anyway.
	# But... allow it anyway.
	CAN_INDEX = True

	def _fromStorage(self, value):
		if not value:
			value = ''
		else:
			value = to_unicode(value)

		try:
			pks = [int(x) for x in value.split(',')]
		except Exception as e:
			raise ValueError('Got exception pulling multi link field from storage:  %s:   %s\n' %(e.__class__.__name__, str(e)))

		return ForeignLinkMultiData(pk=pks, foreignModel=self.foreignModel)

	def _fromInput(self, values):
		if issubclass(values.__class__, (tuple, list, set)):
			pks = []
			objs = []
			for value in values:
				if hasattr(value, '_is_ir_model'):

					pks.append(value._id)
					objs.append(value)

				elif isinstance(value, int):

					pks.append(value)
					objs.append(None)

				elif isBaseStringy(value) and value.isdigit():

					pks.append( int(value) )
					objs.append( None )

				elif value in (irNull, None):
					continue
				else:
					raise ValueError('Unknown element in list:  <%s>  %s' %(value.__class__.__name__, repr(value)) )

			return ForeignLinkMultiData( pk=pks, foreignModel=self.foreignModel, obj=objs)

		elif values in (None, irNull):
			return irNull
		elif issubclass(values.__class__, ForeignLinkMultiData):
			return values

		raise ValueError('Unknown input: <%s>   %s' %(values.__class__.__name__, repr(values)) )
	
	def _toStorage(self, value):
		if isinstance(value, ForeignLinkMultiData):
			return ','.join([str(eachPk) for eachPk in value.getPk()])

		elif issubclass(value.__class__, (list, tuple, set)):
			ret = []
			for val in value:
				if hasattr(val, '_is_ir_model'):
					ret.append(str(val._id))
				elif isinstance(val, int):
					ret.append(str(val))
				elif isStringy(val):
					ret.append(str(int(val)))
				else:
					raise ValueError('Unknown element in list:  <%s>  %s' %(value.__class__.__name__, repr(value)) )
			return ','.join(ret)
		elif isBaseStringy(value):
			# For index
			return value


		raise ValueError('Unknown value type headed for storage:   <%s>   %s' %(value.__class__.__name__, repr(value)))
	

	def isMulti(self):
		'''
			isMulti - Returns True if this is a MultiLink object (expects lists), otherwise False (expects object)

			@return <bool>
		'''
		return True

	def _toIndex(self, value):
		# Support passing either an integer or the model itself
		if issubclass(value.__class__, IRForeignMultiLinkField):
			return ','.join(value.getPk())

		return super(IRForeignLinkField, self)._toIndex(value)


# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestCopyModel - Copy Model test
#

# vim: set ts=4 sw=4 expandtab

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

import copy
import sys
import subprocess
from IndexedRedis import IndexedRedisModel, IRField, InvalidModelException, validatedModels
from IndexedRedis.fields import IRPickleField

# vim: ts=4 sw=4 expandtab

class TestCopyModel(object):

    KEEP_DATA = False

    def setup_method(self, testMethod):
     
        self.model = None

        if testMethod == self.test_copyModel:
            class TestCopyModel_CopyModel(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRField('num', valueType=int, defaultValue=3),
                ]

                INDEXED_FIELDS = ['num']

                KEY_NAME = 'IRTestCopyModel__CopyModel'
            
            self.model = TestCopyModel_CopyModel

        elif testMethod in (self.test_copyInstance, self.test_deepcopyInstance):
            class TestCopyModel_CopyInstance(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRPickleField('tags'),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'IRTestCopyModel__CopyInstance'

            self.model = TestCopyModel_CopyInstance

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()


    def test_copyModel(self):

        Model = self.model

        obj = Model()

        assert obj.num == 3 , 'Checking that original defaultValue=3 is applied'

        ModelCopy = Model.copyModel()

        assert id(Model) != id(ModelCopy) , 'Expected copy of model not to have same id'
        assert Model.__name__ != ModelCopy.__name__ , 'Expected copy of model to have different name'

        assert id(Model.INDEXED_FIELDS) != id(ModelCopy.INDEXED_FIELDS) , 'Expected INDEXED_FIELDS to have different id on copied model'

        assert id(Model.FIELDS) != id(ModelCopy.FIELDS) , 'Expected FIELDS to have different id on copied model'

        assert id(Model.FIELDS[0]) != id(ModelCopy.FIELDS[0]) , 'Expected individual entries in FIELDS to have different id in copy'

        assert ModelCopy.INDEXED_FIELDS == ['num'] , 'Expected INDEXED_FIELDS data to be copied'

        assert ModelCopy.KEY_NAME == Model.KEY_NAME , 'Expected KEY_NAME to be the same between original and copy'

        ModelCopy.FIELDS['num'].defaultValue = 17

        obj2 = ModelCopy()

        assert obj2.num == 17 , 'Expected after changing num.defaultValue to 17 on ModelCopy, that instances of ModelCopy get that default'

        obj = Model()

        assert obj.num == 3 , 'Expected after changing num.defaultValue to 17 on ModelCopy, that instances of Model get the original default'

        
        obj.name = 'm1'
        obj2.name = 'm2'

        ids = obj.save()
        assert ids and ids[0] , 'Failed to save on original model'
        ids = obj2.save()
        assert ids and ids[0] , 'Failed to save on copied model'

        objFetched1 = Model.objects.filter(num=17).first()

        assert objFetched1 , 'Expected to be able to fetch object saved on copy with original'

        objFetched2 = ModelCopy.objects.filter(num=3).first()

        assert objFetched2 , 'Expected to be able to fetch object saved on original with copy'


    def test_copyInstance(self):
 
        Model = self.model

        x = Model(name='one', tags=['one', 'two'])

        xcpy = copy.copy(x)

        assert x.hasSameValues(xcpy) , 'Expected hasSameValues to be True on a copy.copy'

        assert not xcpy._id  , 'Expecteed copy.copy not to be linked to the database'

        linkedCopy = x.copy(True)

        assert linkedCopy.hasSameValues(x) , 'Expected hasSameValues to be True on a copy(True)'
        assert linkedCopy._id == x._id , 'Expected id to be copied with copy(True)'

        assert id(linkedCopy.tags) == id(x.tags) , 'Expected copy(copyValues=False) to retain same reference on a list'

        ids = x.save()
        assert ids and ids[0] , 'Failed to save model'

        fetchedObj = Model.objects.filter(name='one').first()

        assert fetchedObj , 'Failed to fetch object'


        obj = fetchedObj

        objCopy = copy.copy(fetchedObj)

        assert objCopy.hasSameValues(obj) , 'Expected after fetch copy to have same values as original'

        assert id(obj.tags) == id(objCopy.tags) , 'Expected copy.copy to not deepcopy the values, so a list copied should share same reference.'

    def test_deepcopyInstance(self):
        Model = self.model

        x = Model(name='one', tags=['one', 'two'])

        xcpy = copy.deepcopy(x)

        assert x.hasSameValues(xcpy) , 'Expected hasSameValues to be True on a copy.deepcopy'

        assert not xcpy._id  , 'Expecteed copy.deepcopy not to be linked to the database'

        linkedCopy = x.copy(True, copyValues=True)

        assert linkedCopy.hasSameValues(x) , 'Expected hasSameValues to be True on a copy(True, True)'
        assert linkedCopy._id == x._id , 'Expected id to be copied with copy(True, True)'
        assert id(linkedCopy.tags) != id(x.tags) , 'Expected copy(copyValues=True) to NOT retain same reference on a list'
        assert linkedCopy.tags == x.tags , 'Expected list to contain same values with copy(copyValues=True)'

        ids = x.save()
        assert ids and ids[0] , 'Failed to save model'

        fetchedObj = Model.objects.filter(name='one').first()

        assert fetchedObj , 'Failed to fetch object'


        obj = fetchedObj

        objCopy = copy.deepcopy(fetchedObj)

        assert objCopy.hasSameValues(obj) , 'Expected after fetch copy to have same values as original'

        assert id(obj.tags) != id(objCopy.tags) , 'Expected copy.deepcopy to deepcopy the values, so a list copied should share same reference.'

            
if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRRawField - Test default values for fields
#

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

# vim: set ts=4 sw=4 st=4 expandtab

import sys
import subprocess

from IndexedRedis import IndexedRedisModel, irNull
from IndexedRedis.compat_str import tobytes
from IndexedRedis.fields import IRRawField, IRField

# vim: ts=4 sw=4 expandtab

class TestIRRawField(object):
    '''
        TestIRRawField - Test IRRawField
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        if testMethod in (self.test_general, self.test_variousTypes):
            class Model_RawValue(IndexedRedisModel):
                
                FIELDS = [
                    IRField('name'),
                    IRRawField('value'),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME='TestIRRawField__ModelRawValue'

            self.model = Model_RawValue

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()


    def test_general(self):
        
        Model = self.model

        obj = Model()

        updatedFields = obj.getUpdatedFields()

        helloWorldRaw = b'\x01Hello World\x01'

        assert updatedFields == {} , 'Expected no updated fields when object is first created.\nExpected: %s\nGot:     %s' %(repr([]), repr(updatedFields) )

        assert obj.value == '', 'Expected default value of IRRawField to be empty string'

        obj.name = 'one'

        obj.save()

        assert obj.getUpdatedFields() == {} , 'Expected no updated fields after object is saved'

        obj.value = u'Goodbye'

        assert obj.value == u'Goodbye' , 'Expected IRRawField value to be u"goodbye" after setting to u"goodbye" (no conversion)'

        obj.value = helloWorldRaw

        assert obj.value == helloWorldRaw , 'Expected IRRawField value to be b"\\x01Hello World\\x01" after setting to "\\x01Hello World\\x01"'

        
        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == helloWorldRaw, 'Expected asDict(forStorage=False) to contain IRRawField value as bytes string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == helloWorldRaw, 'Expected asDict(forStorage=True) to contain IRRawField value as bytes string. \nExpected: %s\nGot:     %s' %(repr(helloWorldRaw), repr(dictForStorage['value']) )

        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating'

        assert updatedFields['value'][0] ==  '', 'Expected old value to be empty string in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == helloWorldRaw , 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving.'

        fetchObj = Model.objects.filter(name='one').first()


        assert fetchObj , 'Expected to be able to fetch object on name="one" after saving.'

        obj = fetchObj

        assert obj.value == helloWorldRaw , 'Expected value of fetched to be converted string (As bytes), %s. Got: %s' %(repr(helloWorldRaw), repr(fetchObj.value), )

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after fetching'

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == helloWorldRaw, 'After fetching, Expected asDict(forStorage=False) to contain IRRawField value as bytes string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == helloWorldRaw, 'After fetching, Expected asDict(forStorage=True) to contain IRRawField value as bytes string. \nExpected: %s\nGot:     %s' %(repr(helloWorldRaw), repr(dictForStorage['value']) )

        obj.value = b'q123'


        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating on fetched object'

        assert updatedFields['value'][0] == helloWorldRaw , 'Expected old value to be %s in updatedFields. Got: %s' %(repr(helloWorldRaw), repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == b'q123' , 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == b'q123', 'After fetching, then updating, Expected asDict(forStorage=False) to contain IRRawField value as bytes string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == b'q123', 'After fetching, then updating, Expected asDict(forStorage=True) to contain IRRawField value as bytes string.\nExpected: %s\nGot:     %s' %(repr(b'q123'), repr(dictForStorage['value']) )

        
        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving'


    def test_variousTypes(self):

        Model = self.model

        obj = Model()

        obj.name = 'two'
        obj.value = 5

        assert obj.value == 5 , 'Expected value when set to int(5) to remain int(5)'


        ids = obj.save()

        assert ids and ids[0] , 'Failed to save with integer value'

        fetchedObj = obj.objects.filter(name='two').first()

        assert fetchedObj , 'Failed to fetch object'

        obj = fetchedObj

        assert obj.value == b'5' , 'Expected value to be raw b"5" from redis'

        obj.value = 'abc'

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save with string value'

        fetchedObj = obj.objects.filter(name='two').first()
        assert fetchedObj , 'Failed to fetch object'

        obj = fetchedObj

        assert obj.value == b'abc' , 'Expected string value to be retrieved as bytes'

        obj.value = 3.14

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save with float value'

        fetchedObj = obj.objects.filter(name='two').first()
        assert fetchedObj , 'Failed to fetch object'

        obj = fetchedObj

        assert type(obj.value) == bytes , 'Expected to fetch saved float as bytes'

        assert obj.value == b'3.14' , 'Expected fetched save float to be b"3.14"'

        assert float(obj.value) == 3.14 , 'Expected float(obj.value) to be as originally saved'

        obj.value = irNull

        ids = obj.save()
        assert ids and ids[0] , 'Failed to save with float value'

        fetchedObj = obj.objects.filter(name='two').first()
        assert fetchedObj , 'Failed to fetch object'

        obj = fetchedObj

        assert obj.value == b'' , 'Expected irNull to be converted to empty string (since no conversion). Got: %s' %( repr(obj.value), )




if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



#!/usr/bin/env python

import datetime
import sys
import IndexedRedis
from IndexedRedis import IndexedRedisModel, IRField
from IndexedRedis.fields import IRRawField

# vim: ts=4 sw=4 expandtab

class MyRawModel(IndexedRedisModel):

    FIELDS = [ \
        IRField('name'),
        IRRawField('data'),
    ]

    INDEXED_FIELDS = [ \
        'name',
#        'data',
    ]

    KEY_NAME = 'MyRawModel'


if __name__ == '__main__':

    MyRawModel.objects.delete()
    
    newObj = MyRawModel(name='hello', data=b'hello world')

    print ( "My data: <%s>%s\n\n" %(str(type(newObj.data)), repr(newObj.data)))
    print ( "My dict (not for storage): %s\n\n" %(str(newObj.asDict(forStorage=False, strKeys=True)),))
    print ( "My dict (for storage): %s\n\n" %(str(newObj.asDict(forStorage=True, strKeys=True)),))

    newObj.save()

    fetch1 = MyRawModel.objects.filter(name='hello').first()

    if not fetch1:
        print ( "No result for fetch1\n\n" )
    else:
        print ( "fetch1 data: <%s>%s\n\n" %(str(type(fetch1.data)), repr(fetch1.data)))
        print ( "fetch1 dict (not for storage): %s\n\n" %(str(fetch1.asDict(forStorage=False, strKeys=True)),))
        print ( "fetch1 dict (for storage): %s\n\n" %(str(fetch1.asDict(forStorage=True, strKeys=True)),))
        
#    fetch2 = MyRawModel.objects.filter(data=datetime.datetime(1989, 6, 28, 12, 12, 0)).first()
#    if not fetch2:
#        print ( "No result for fetch2\n\n" )
#    else:
#        print ( "fetch2 data: <%s>%s\n\n" %(str(type(fetch2.data)), repr(fetch2.data)))
#        print ( "fetch2 dict (not for storage): %s\n\n" %(str(fetch2.asDict(forStorage=False, strKeys=True)),))
#        print ( "fetch2 dict (for storage): %s\n\n" %(str(fetch2.asDict(forStorage=True, strKeys=True)),))



#!/usr/bin/env python

import datetime
import sys
import IndexedRedis
from IndexedRedis import IndexedRedisModel, IRField, setDefaultIREncoding
from IndexedRedis.fields import IRCompressedField, IRUnicodeField, IRFieldChain, IRBase64Field


# Make sure native type will fail
setDefaultIREncoding('ascii')

# on python2 you'll probably get some unicode escape sequences, but you'll get pretty pictures on python3
prettyPictures = b' \xe2\x9c\x8f \xe2\x9c\x90 \xe2\x9c\x91 \xe2\x9c\x92 \xe2\x9c\x93 \xe2\x9c\x94 \xe2\x9c\x95 \xe2\x9c\x96 \xe2\x9c\x97 \xe2\x9c\x98 \xe2\x9c\x99 \xe2\x9c\x9a \xe2\x9c\x9b \xe2\x9c\x9c \xe2\x9c\x9d \xe2\x9c\x9e \xe2\x9c\x9f \xe2\x9c\xa0 \xe2\x9c\xa1 \xe2\x9c\xa2 \xe2\x9c\xa3 \xe2\x9c\xa4 \xe2\x9c\xa5 \xe2\x9c\xa6 \xe2\x9c\xa7 \xe2\x9c\xa9 \xe2\x9c\xaa \xe2\x9c\xab '

# vim: ts=4 sw=4 expandtab

class MyUnicodeModel(IndexedRedisModel):

    FIELDS = [ \
        IRField('name'),
        IRFieldChain('unicodeField', [IRUnicodeField(encoding='utf-8'), IRBase64Field(encoding='utf-8')]),
    ]

    INDEXED_FIELDS = [ \
        'name',
#        'data',
    ]

    KEY_NAME = 'MyUnicodeModel'


if __name__ == '__main__':

    MyUnicodeModel.objects.delete()
    newObj = MyUnicodeModel(name='hello', unicodeField=prettyPictures)

    print ( "My unicodeField: <%s>%s\n\n" %(str(type(newObj.unicodeField)), repr(newObj.unicodeField)))
    print ( "My dict (not for storage): %s\n\n" %(str(newObj.asDict(forStorage=False, strKeys=True)),))
    print ( "My dict (for storage): %s\n\n" %(str(newObj.asDict(forStorage=True, strKeys=True)),))

    newObj.save()

    fetch1 = MyUnicodeModel.objects.filter(name='hello').first()

    if not fetch1:
        print ( "No result for fetch1\n\n" )
    else:
        print ( "fetch1 unicodeField: <%s>%s\n\n" %(str(type(fetch1.unicodeField)), repr(fetch1.unicodeField)))
        print ( "fetch1 dict (not for storage): %s\n\n" %(str(fetch1.asDict(forStorage=False, strKeys=True)),))
        print ( "fetch1 dict (for storage): %s\n\n" %(str(fetch1.asDict(forStorage=True, strKeys=True)),))
        
#    fetch2 = MyUnicodeModel.objects.filter(data=datetime.datetime(1989, 6, 28, 12, 12, 0)).first()
#    if not fetch2:
#        print ( "No result for fetch2\n\n" )
#    else:
#        print ( "fetch2 data: <%s>%s\n\n" %(str(type(fetch2.data)), repr(fetch2.data)))
#        print ( "fetch2 dict (not for storage): %s\n\n" %(str(fetch2.asDict(forStorage=False, strKeys=True)),))
#        print ( "fetch2 dict (for storage): %s\n\n" %(str(fetch2.asDict(forStorage=True, strKeys=True)),))



# Copyright (c) 2014, 2015, 2016, 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# chain - Support for chaining multiple IRField's (so for example to compress a json value)
#


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

from . import IRField, irNull


class IRFieldChain(IRField):
	'''
	   IRFieldChain -
	   
	   These chain together the operations from multiple fields.

	   toStorage is applied left-to-right,
	   fromInput and fromStorage is applied right-to-left.

	   The output of one field is the input of the next.

	   An IRFieldChain is indexable so long as ALL of its contained fields are indexable.
	   An IRFieldChain is forced to be a hashed index if the rightmost field forces it (as some types do),
	     otherwise hashIndex will determine if it is hashed.
	'''

	# CAN_INDEX will be determined if the right-most field in chainedFields is indeable.
	#   hashIndex will also be forced if the right-most field forces it.
	CAN_INDEX = False

	def __init__(self, name, chainedFields, defaultValue=irNull, hashIndex=False):
		'''
			__init__ - Create an IRFieldChain object.

			  These chain together the operations from multiple fields.

			  toStorage is applied left-to-right,
			  fromInput and fromStorage is applied right-to-left.

			  The output of one field is the input of the next.

			   An IRFieldChain is indexable so long as ALL of its contained fields are indexable.
			   An IRFieldChain is forced to be a hashed index if the rightmost field forces it (as some types do),
			     otherwise hashIndex will determine if it is hashed.

			@see IRField.__init__
		'''
		if not name:
			raise ValueError('IRFieldChain has empty name.')

		self.chainedFields = []

		if not chainedFields:
			raise ValueError('IRFieldChain has no chained fields defined!')

		for field in chainedFields:
			# If we got jsut a class, construct it.
			if type(field) == type:
				field = field()
			if str(field) != '':
				raise ValueError('IRFieldChain has chained fields with a name set. The name should only be provided to the IRFieldChain object itself.')
			if not issubclass(field.__class__, IRField):
				raise ValueError("IRFieldChain's 'chainedFields' (second arg) should only contain IRField objects or subclasses thereof. Got: %s %s" %(str(type(field)), repr(field)))


			self.chainedFields.append(field)

		self.defaultValue = defaultValue

		(canIndex, forceHashIndex) = self._checkCanIndex()

		self.CAN_INDEX = canIndex

		if hashIndex is False:
			hashIndex = forceHashIndex

		self.hashIndex = hashIndex


	def _toStorage(self, value):
		'''
			_toStorage - Convert the value to a string representation for storage.

			@param value - The value of the item to convert
			@return A string value suitable for storing.
		'''

		for chainedField in self.chainedFields:
			value = chainedField.toStorage(value)

		return value

	def _fromStorage(self, value):
		'''
			_fromStorage - Convert the value from storage (string) to the value type.

			@return - The converted value, or "irNull" if no value was defined (and field type is not default/string)
		'''

		for chainedField in reversed(self.chainedFields):
			value = chainedField._fromStorage(value)

		return value
	
	def _fromInput(self, value):

		for chainedField in reversed(self.chainedFields):
			value = chainedField._fromInput(value)

		return value

	def _toIndex(self, value):
		
		for chainedField in self.chainedFields:
			value = chainedField._toIndex(value)

		return value

	def _getReprProperties(self):
		chainedFieldsRepr = []
		for chainedField in self.chainedFields:
			fieldRepr = chainedField.__class__.__name__ + '( ' + ', '.join(chainedField._getReprProperties()) + ' )'
			chainedFieldsRepr.append(fieldRepr)

		return [
			'chainedFields=[ %s ]' %(', '.join(chainedFieldsRepr), ),
			'hashIndex=%s' %(str(self.hashIndex),)
		]

	def _checkCanIndex(self):
		'''
			_checkCanIndex - Check if we CAN index (if all fields are indexable).
				Also checks the right-most field for "hashIndex" - if it needs to hash we will hash.
		'''

		# NOTE: We can't just check the right-most field. For types like pickle that don't support indexing, they don't
		#   support it because python2 and python3 have different results for pickle.dumps on the same object.
		#   So if we have a field chain like Pickle, Compressed   then we will have two different results.
		if not self.chainedFields:
			return (False, False)

		for chainedField in self.chainedFields:
			if chainedField.CAN_INDEX is False:
				return (False, False)

		return (True, self.chainedFields[-1].hashIndex)

	def copy(self):
		return self.__class__(name=self.name, chainedFields=[field.copy() for field in self.chainedFields], defaultValue=self.defaultValue)


	def __new__(self, name, chainedFields=None, defaultValue=irNull, hashIndex=False):
		if not name:
			raise ValueError('IRChainedField defined without a name!')

		return IRField.__new__(self, name)


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



# Copyright (c) 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# deprecated - Mark things as deprecated.
#

# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

import time
import sys

__all__ = ('funcTimeMap', 'timeFunc', 'getFuncTimes')

global funcTimeMap
funcTimeMap = {}


def timeFunc(func):

	if func not in funcTimeMap:
		funcTimeMap[func] = 0.0

	def _timeFunc_wrapper(*args, **kwargs):
		t1 = time.time()
		_exc = None
		try:
			ret = func(*args, **kwargs)
			t2 = time.time()
		except Exception as e:
			t2 = time.time()
			_exc = e

		funcTimeMap[func] += (t2 - t1)
		if _exc is not None:
			raise _exc

		return ret
	return _timeFunc_wrapper


def getFuncTimes():
	
	global funcTimeMap

	ret = []

	for func, totalTime in funcTimeMap.items():
		ret.append ( (func.__qualname__, func, totalTime) )

	ret.sort(key = lambda x : x[2])

	ret.reverse()


	return ret

def printTopTimes(funcTimes, numTop=10):

	topTimes = funcTimes[:numTop]

	ret = []

	maxFirstString = max( [len(x[0]) for x in topTimes ] )

	formatStr = "%-" + str(maxFirstString + 1) + "s    "

	headerLine = (formatStr + "%s") % ( "Func Name", "Total Time" )

	print (headerLine )
	print ( "-" * len(headerLine) )
	print ( "" )

	formatStr += "%.5f"

	for item in topTimes:
		print ( formatStr % (item[0], item[2]) )
		

# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRFixedPointField - Test default values for fields
#

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

# vim: set ts=4 sw=4 st=4 expandtab

import copy
import sys
import subprocess

from IndexedRedis import IndexedRedisModel, irNull
from IndexedRedis.compat_str import tobytes
from IndexedRedis.fields import IRFixedPointField, IRField

# vim: ts=4 sw=4 expandtab

class TestIRFixedPointField(object):
    '''
        TestIRFixedPointField - Test IRFixedPointField
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        if testMethod in (self.test_general, ):
            class Model_FixedPointValue(IndexedRedisModel):
                
                FIELDS = [
                    IRField('name'),
                    IRFixedPointField('value', defaultValue=irNull),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME='TestIRFixedPointField__ModelFixedPointValue'

            self.model = Model_FixedPointValue
        elif testMethod in (self.test_defaultValue, ):
            class Model_FixedPointDefaultValue(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRFixedPointField('value', defaultValue=9.91775),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRFixedPointField__ModelFixedPointDefaultValue'

            self.model = Model_FixedPointDefaultValue

        elif testMethod in (self.test_indexFixedPoint, ):
            class Model_IndexFixedPoint(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRFixedPointField('value', defaultValue=irNull),
                    IRFixedPointField('value2', defaultValue=1.11),
                ]

                INDEXED_FIELDS = ['name', 'value', 'value2']

                KEY_NAME = 'TestIRFixedPointField__IndexFixedPoint'

            self.model = Model_IndexFixedPoint

        elif testMethod == self.test_decimalPlaces:
            class Model_DecimalPlaces(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRFixedPointField('value1', decimalPlaces=1),
                    IRFixedPointField('value4', decimalPlaces=4),
                    IRFixedPointField('value9', decimalPlaces=9),
                    IRFixedPointField('value4MoreDef', decimalPlaces=4, defaultValue=5.123456),
                ]

                INDEXED_FIELDS = ['name', 'value1', 'value4', 'value9']

                KEY_NAME = 'TestIRFixedPointField__DecimalPlaces'

            self.model = Model_DecimalPlaces

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    @staticmethod
    def asFloatStr(val, decimalPlaces):
        formatStr = "%." + str(decimalPlaces) + "f"

        return formatStr % (val, )

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()


    def test_general(self):
        
        Model = self.model

        obj = Model()

        updatedFields = obj.getUpdatedFields()

        floatVal = 9.17256
        numDecimalPlaces = obj.FIELDS[obj.FIELDS.index('value')].decimalPlaces
        floatStr = self.asFloatStr(floatVal, numDecimalPlaces)


        assert updatedFields == {} , 'Expected no updated fields when object is first created.\nExpected: %s\nGot:     %s' %(repr([]), repr(updatedFields) )

        assert obj.value == irNull , 'Expected default value of IRFixedPointField to be irNull when defaultValue=irNull'

        obj.name = 'one'

        obj.save()

        assert obj.getUpdatedFields() == {} , 'Expected no updated fields after object is saved'


        obj.value = '5.66'

        assert obj.value == 5.66 , 'Expected setting value as string to be converted to float'

        obj.value = floatVal

        assert obj.value == floatVal, 'Expected IRFixedPointField value to be the float value it was set to'

        
        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))


        assert dictConverted['value'] == floatVal, 'Expected asDict(forStorage=False) to contain value as float Got: %s' %( repr(dictConverted['value']), )
        assert dictForStorage['value'] == floatStr, 'Expected asDict(forStorage=True) to contain IRFixedPointField value fixed string \nExpected: %s\nGot:     %s' %(repr(floatStr), repr(dictForStorage['value']) )

        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating'

        assert updatedFields['value'][0] == irNull , 'Expected old value to be irNull in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] ==  floatVal, 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving.'

        fetchObj = Model.objects.filter(name='one').first()


        assert fetchObj , 'Expected to be able to fetch object on name="one" after saving.'

        obj = fetchObj

        assert obj.value == floatVal, 'Expected value of fetched to be float value %s. Got: %s' %(repr(floatVal), repr(fetchObj.value), )

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after fetching'

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == floatVal, 'After fetching, Expected asDict(forStorage=False) to contain IRFixedPointField value as float Got: %s' %( repr(dictConverted['value']), )
        assert dictForStorage['value'] == floatStr, 'After fetching, Expected asDict(forStorage=True) to contain IRFixedPointField value as fixed string. \nExpected: %s\nGot:     %s' %(repr(floatStr), repr(dictForStorage['value']) )


    def test_defaultValue(self):

        Model = self.model

        obj = Model()

        valField = obj.FIELDS[obj.FIELDS.index('value')]

        numDecimalPlaces = valField.decimalPlaces
        defaultValue = 9.91775
        defaultValueStr = self.asFloatStr(defaultValue, numDecimalPlaces)

        assert valField.defaultValue == defaultValue , 'Expected defaultValue attribute to be set'

        assert obj.value == defaultValue , 'Expected defaultValue to be applied to a IRFixedPointField field.\nExpected: %f\nGot:     %s' %(defaultValue, repr(obj.value), )

        obj.name = 'test'

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save value'

        assert obj.value == defaultValue , 'Expected defaultValue to remain on a fixed point field after saving'

        objFetched = Model.objects.filter(name='test').first()
        assert objFetched , 'Expected to be able to fetch object'

        obj = objFetched

        assert obj.value == defaultValue , 'Expected defaultValue to remain on a fixed point field after fetching'

        floatVal = 3.14
        floatValStr = self.asFloatStr(floatVal, numDecimalPlaces)

        obj.value = floatVal

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save object'

        objFetched = Model.objects.filter(name='test').first()
        assert objFetched , 'Expected to be able to fetch object'

        obj = objFetched

        assert obj.value == floatVal , 'Expected to be able to change value from default.'

    def test_indexFixedPoint(self):

        Model = self.model

        obj = Model()

        obj.name = 'test'

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save object'

        floatVal = 2.61
        numDecimalPlaces = obj.FIELDS[obj.FIELDS.index('value')].decimalPlaces
        floatStr = self.asFloatStr(floatVal, numDecimalPlaces)

        otherFloatVal = 5.66
        otherNumDecimalPlaces = obj.FIELDS[obj.FIELDS.index('value')].decimalPlaces
        otherFloatStr = self.asFloatStr(floatVal, numDecimalPlaces)

        # Add a "distraction" object
        otherObj = Model()
        otherObj.value = floatVal
        otherObj.value2 = otherFloatVal

        otherObj.save()

        objFetched = Model.objects.filter(value=irNull).first()

        assert objFetched , 'Failed to fetch an object with default value (irNull) on a fixed point field'

        d1 = obj.asDict(includeMeta=True, forStorage=False, strKeys=True)
        d2 = objFetched.asDict(includeMeta=True, forStorage=False, strKeys=True)

        assert objFetched == obj , 'Object fetched was not expected object.\nExpected: %s\n\nGot:     %s\n' %( 
                obj.asDict(includeMeta=True, forStorage=False, strKeys=True),
                objFetched.asDict(includeMeta=True, forStorage=False, strKeys=True)
                )

        objFetched = Model.objects.filter(value2=otherFloatVal).first()

        assert objFetched , 'Failed to fetch object using a non-default value on fixed point field'

        assert objFetched == otherObj , 'Object fetched was not expected object.\nExpected: %s\n\nGot:     %s\n' %( 
                otherObj.asDict(includeMeta=True, forStorage=False, strKeys=True),
                objFetched.asDict(includeMeta=True, forStorage=False, strKeys=True)
                )

    def test_decimalPlaces(self):
        
        #IRFixedPointField('value1', decimalPlaces=1),
        #IRFixedPointField('value4', decimalPlaces=4),
        #IRFixedPointField('value9', decimalPlaces=9),
        #IRFixedPointField('value4MoreDef', decimalPlaces=4, defaultValue=5.123456),

        Model = self.model

        obj = Model()

        assert obj.value4MoreDef == 5.1235 , 'Expected when default value has more decimal places than the field supports, for it to be rounded.\nExpected: 5.1235\nGot:     %f' %(obj.value4MoreDef, )

        obj.value1 = '1.41'

        assert obj.value1 == 1.4 , 'Expected "1.41" to be converted to float(1.4) when decimalPlaces=1. Got: %s' %(repr(obj.value1, ))

        obj.value1 = '1.48'

        assert obj.value1 == 1.5 , 'Expected "1.48" to be converted to float(1.48) when decimalPlaces=1. Got: %s' %(repr(obj.value1, ))

        obj = Model(value1='1.48')

        assert obj.value1 == 1.5 , 'Expected "1.48" to be converted to float(1.48) when decimalPlaces=1 (constructor). Got: %s' %(repr(obj.value1, ))


        val = '99145.3362816911682'

        valFloat = float(val)
        rvalFloat1 = round(valFloat, 1)
        floatStr1 = self.asFloatStr(valFloat, 1)
        rvalFloat4 = round(valFloat, 4)
        floatStr4 = self.asFloatStr(valFloat, 4)
        rvalFloat9 = round(valFloat, 9)
        floatStr9 = self.asFloatStr(valFloat, 9)
 
        obj.value1 = valFloat
        obj.value4 = valFloat
        obj.value9 = valFloat


        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))


        assert dictConverted['value1'] == rvalFloat1 , 'Expected converted value1 to be %f. Got %f. ' %(rvalFloat1, dictConverted['value1'])
        assert dictConverted['value4'] == rvalFloat4 , 'Expected converted value4 to be %f. Got %f. ' %(rvalFloat4, dictConverted['value4'])
        assert dictConverted['value9'] == rvalFloat9 , 'Expected converted value9 to be %f. Got %f. ' %(rvalFloat9, dictConverted['value9'])

        assert dictForStorage['value1'] == floatStr1 , 'Expected storage value1 to be %s. Got %s. ' %( floatStr1, dictForStorage['value1'] )
        assert dictForStorage['value4'] == floatStr4 , 'Expected storage value4 to be %s. Got %s. ' %( floatStr4, dictForStorage['value4'] )
        assert dictForStorage['value9'] == floatStr9 , 'Expected storage value9 to be %s. Got %s. ' %( floatStr9, dictForStorage['value9'] )


        ids = obj.save()

        assert ids and ids[0] , 'Failed to save object'

        objFetched = Model.objects.filter(value1=rvalFloat1).first()
        assert objFetched , 'Failed to fetch object with exact value for value1'
        objFetched = Model.objects.filter(value4=rvalFloat4).first()
        assert objFetched , 'Failed to fetch object with exact value for value4'
        objFetched = Model.objects.filter(value9=rvalFloat9).first()
        assert objFetched , 'Failed to fetch object with exact value for value9'


        obj = objFetched

        assert obj.value1 == rvalFloat1 , 'Expected value1 to be correct after fetching. Expected: %f Got: %f' %(rvalFloat1, obj.value1)
        assert obj.value4 == rvalFloat4 , 'Expected value4 to be correct after fetching. Expected: %f Got: %f' %(rvalFloat4, obj.value4)
        assert obj.value9 == rvalFloat9 , 'Expected value9 to be correct after fetching. Expected: %f Got: %f' %(rvalFloat9, obj.value9)

        objFetched = Model.objects.filter(value1=rvalFloat9).first()
        assert objFetched , 'Expected to be able to fetch using a value of higher precision with same rounded-value as index. Failed on dec=1'
        objFetched = Model.objects.filter(value4=rvalFloat9).first()
        assert objFetched , 'Expected to be able to fetch using a value of higher precision with same rounded-value as index. Failed on dec=4'

        objFetched = Model.objects.filter(value4=val).first()
        assert objFetched , 'Expected to be able to fetch using a string of higher precision with same rounded-value as index.'

        objFetched = Model.objects.filter(value9=rvalFloat4).first()
        assert not objFetched , 'Expected to NOT be able to fetch object using a lower-precision value than the index, where those two are different values.'

        OrigModel = Model.copyModel()

        ## NOTE - model is CHANGED AFTER THIS POINT ##

        index = "%.5f" % (round(float(Model.objects.first().value4), 5), )

        # Increase precision by 1 on value4
        Model.FIELDS[Model.FIELDS.index('value4')].decimalPlaces = 5
        
#        obj = Model.objects.first()

        forStorageDict = obj.asDict(forStorage=True)

        floatStr5 = self.asFloatStr(rvalFloat4, 5)
        assert forStorageDict['value4'] == floatStr5 , 'TEST expected that increasing decimal places on model field would take affect. It did not?'

        # Should not work as different fixed point representation:
        objFetched = Model.objects.filter(value4=val).first()
        assert not objFetched , 'Expected NOT to still be able to fetch on index of higher precision, when value changes, after decimalPlaces was increased since save'

        objFetched = Model.objects.filter(value4=rvalFloat4).first()
        assert not objFetched , 'Expected NOT to still be able to fetch an index of former precision, when value rounds to the same float but DIFFERENT fixed-point (different final-2 places) after decimalPlaces was increased since save'

        objFetched = Model.objects.filter(value4=rvalFloat9).first()
        assert not objFetched ,'Expected NOT to still be able to fetch an index of former precision, when value rounds to the same float but DIFFERENT fixed-point (extra 0s) after decimalPlaces was increased since save'

        # NOTE: Reindex will leave this field indexed on two different values for the field that changed -
        #   So we must use reset instead of Model.objects.reindex()
        Model.reset(Model.objects.all())

        # Now index is .33630 - because it was originally rounded out 4 from val, no longer will match round(val, 5) which is .33628

        newVal = '99145.3363012'

        objFetched = Model.objects.filter(value4=99145.33630).first()
        assert objFetched , 'After reindex, expected to be able to fetch on new index'

        objFetched = OrigModel.objects.filter(value4=99145.3363).first()
        assert not objFetched , 'After reindex, expected NOT to be able to fetch on the old index'


if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields.classic - The IRField type which behaves like the "classic" IndexedRedis string-named fields.
#


# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

from . import IRField

class IRClassicField(IRField):
	'''
		IRClassicField - The IRField type which behaves like the "classic" IndexedRedis string-named fields.
		

		This will store and retrieve data encoding into the default encoding (@see IndexedRedis.compat_str.setDefaultIREncoding)

		and have a default value of empty string.

		An IRClassicField is indexable, but there is no option to hash the index.
	'''

	CAN_INDEX = True

	def __init__(self, name=''):
		'''
			__init__ - Create an IRClassicField.
			  This field behaves as a plain string entry in FIELDS did prior to 5.0.0.

			  The default value is empty string, and it stores strings using defaultIREncoding.

			  Consider changing to an IRField or one of the other many types.

			  @param name <str> - Field name
		'''
		IRField.__init__(self, name=name, valueType=str, defaultValue='')

	def __new__(self, name=''):
		return IRField.__new__(self, name)

	def copy(self):
		return self.__class__(name=self.name)


# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



# Copyright (c) 2014, 2015, 2016, 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# IRQueryableList - QueryableList with some added callbacks to IndexedRedis
#

# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

import sys

__all__ = ('defaultIREncoding', 'to_unicode', 'tobytes', 'setDefaultIREncoding', 'getDefaultIREncoding')

try:
	global defaultIREncoding
	defaultIREncoding = sys.getdefaultencoding()
	if defaultIREncoding == 'ascii':
		defaultIREncoding = 'utf-8'
except:
	defaultIREncoding = 'utf-8'

# Encoding stuff

def setDefaultIREncoding(encoding):
	'''
		setDefaultIREncoding - Sets the default encoding used by IndexedRedis.
		  This will be the default encoding used for field data. You can override this on a
		  per-field basis by using an IRField (such as IRUnicodeField or IRRawField)

		@param encoding - An encoding (like utf-8)
	'''
	try:
		b''.decode(encoding)
	except:
		raise ValueError('setDefaultIREncoding was provided an invalid codec. Got (encoding="%s")' %(str(encoding), ))

	global defaultIREncoding
	defaultIREncoding = encoding

def getDefaultIREncoding():
	'''
		getDefaultIREncoding - Get the default encoding that IndexedRedis will use for all field data.
		  You can override this on a per-field basis by using an IRField (such as IRUnicodeField or IRRawField)

		  @return <str> - Default encoding string
	'''
	global defaultIREncoding
	return defaultIREncoding


# String Assurance

if bytes == str:
	# Python 2
	def to_unicode(x, encoding=None):
		'''
			to_unicode - Ensure a given value is decoded into unicode type

			@param x - Value
			@param encoding <None/str> (default None) - None to use defaultIREncoding, otherwise an explicit encoding

			@return - "x" as a unicode type
		'''
		if encoding is None:
			global defaultIREncoding
			encoding = defaultIREncoding

		if isinstance(x, unicode):
			return x
		elif isinstance(x, str):
			return x.decode(encoding)
		else:
			return str(x).decode(encoding)


	def tobytes(x, encoding=None):
		'''
			tobytes - Ensure that a given value is encoded into bytes (str on python2)

			@param x - Value
			@param encoding <None/str> (default None) - None to use defaultIREncoding, otherwise an explicit encoding

			@return - "x" as a bytes (str on python2) type
		'''
		if encoding is None:
			global defaultIREncoding
			encoding = defaultIREncoding

		if isinstance(x, str):
			return x
		try:
			return str(x)
		except:
			return x.encode(encoding)
#	tobytes = lambda x : str(x)

	def isStringy(x):
		'''
			isStringy - Check if a given value extends from any string-like type (basestring on python2)

			@param x - Value to check

			@return <bool> - True if "x" is a stringy type
		'''
		return issubclass(x.__class__, basestring)

	def isBaseStringy(x):
		'''
			isBaseStringy - Chek if is a "basestring" extending class.

			@param x - Value to chek

			@return <bool> - True if "x" is a "basestring" extending type
		'''
		return issubclass(x.__class__, basestring)
	
	encoded_str_type = unicode

	def isEmptyString(x):
		'''
			isEmptyString - Check if x is an empty string
		'''
		if issubclass(x.__class__, str):
			return bool(x == '')
		elif issubclass(x.__class__, unicode):
			return bool(x == u'')

		return False

else:
	# Python 3
	def to_unicode(x, encoding=None):
		'''
			to_unicode - Ensure a given value is decoded into unicode type (str on python3)

			@param x - Value
			@param encoding <None/str> (default None) - None to use defaultIREncoding, otherwise an explicit encoding

			@return - "x" as a unicode (str on python3) type
		'''
		if encoding is None:
			global defaultIREncoding
			encoding = defaultIREncoding

		if isinstance(x, bytes) is False:
			return str(x)
		return x.decode(encoding)

	def tobytes(x, encoding=None):
		'''
			tobytes - Ensure that a given value is encoded into bytes

			@param x - Value
			@param encoding <None/str> (default None) - None to use defaultIREncoding, otherwise an explicit encoding

			@return - "x" as a bytes type
		'''
		if encoding is None:
			global defaultIREncoding
			encoding = defaultIREncoding

		if isinstance(x, bytes) is True:
			return x
		return x.encode(encoding)

	def isStringy(x):
		'''
			isStringy - Check if a given value extends from any string-like type (str or bytes on python3)

			@param x - Value to check

			@return <bool> - True if "x" is a stringy type
		'''
		return issubclass(x.__class__, (str, bytes))

	def isBaseStringy(x):
		'''
			isBaseStringy - Chek if is a "basestring" extending class.

			@param x - Value to chek

			@return <bool> - True if "x" is a "basestring" extending type
		'''
		return issubclass(x.__class__, str)
	

	def isEmptyString(x):
		'''
			isEmptyString - Check if x is an empty string
		'''
		if issubclass(x.__class__, bytes):
			return bool(x == b'')
		elif issubclass(x.__class__, str):
			return bool(x == '')
		return False

	encoded_str_type = str

def isEncodedString(x):
	'''
		isEncodedString - Check if a given string is "encoded" with a codepage.

		  Note this means UNICODE, not BYTES, even though python uses "decode" to apply an encoding and "encode" to get the raw bytes..
	'''
	return issubclass(x.__class__, encoded_str_type)

# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRCompressedField - Test for IRCompressedField
#

# vim: set ts=4 sw=4 st=4 expandtab

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

import sys
import subprocess

import zlib
import bz2

from IndexedRedis import IndexedRedisModel, irNull
from IndexedRedis.compat_str import tobytes
from IndexedRedis.fields import IRCompressedField, IRField

# vim: ts=4 sw=4 expandtab

class TestIRCompressedField(object):
    '''
        TestIRCompressedField - Test IRCompressedField
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        if testMethod == self.test_general:
            class Model_GeneralCompressed(IndexedRedisModel):
                
                FIELDS = [
                    IRField('name'),
                    IRCompressedField('value', defaultValue=irNull),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME='TestIRCompressedField__GeneralCompressed'

            self.model = Model_GeneralCompressed

        elif testMethod == self.test_compressBz2:
            class Model_CompressBz2(IndexedRedisModel):

                FIELDS = [
                    IRField('name'),
                    IRCompressedField('value', compressMode='bz2', defaultValue=irNull),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRCompressedField__CompressBz2'

            self.model = Model_CompressBz2
        elif testMethod == self.test_compressLzma:
            try:
                class Model_CompressLzma(IndexedRedisModel):
                    FIELDS = [
                        IRField('name'),
                        IRCompressedField('value', compressMode='lzma', defaultValue=irNull),
                    ]

                    INDEXED_FIELDS = ['name']

                    KEY_NAME = 'TestIRCompressedField__CompressLzma'
                self.model = Model_CompressLzma
            except ImportError:
                pass

        elif testMethod == self.test_defaultValue:
            class Model_CompressedDefaultValue(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRCompressedField('value', defaultValue=b'woobley'),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME='TestIRCompressedField__CompressedDefaultValue'

            self.model = Model_CompressedDefaultValue

        elif testMethod == self.test_index:
            class Model_CompressedIndex(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRCompressedField('value')
                ]

                INDEXED_FIELDS = ['name', 'value']

                KEY_NAME = 'TestIRCompressedField__CompressedIndex'

            self.model = Model_CompressedIndex

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()


    def test_general(self):
        
        Model = self.model

        obj = Model()

        updatedFields = obj.getUpdatedFields()

        helloWorldBytes = b'\x01Hello World\x01'
        compressedHelloWorldGzip = zlib.compress(helloWorldBytes, 9)

        assert updatedFields == {} , 'Expected no updated fields when object is first created.\nExpected: %s\nGot:     %s' %(repr({}), repr(updatedFields) )

        assert obj.value == irNull , 'Expected default value of IRCompressedField to be irNull when defaultValue=irNull'

        obj.name = 'one'

        obj.save()

        assert obj.getUpdatedFields() == {} , 'Expected no updated fields after object is saved'

        obj.value = u'Goodbye'

        assert obj.value == u'Goodbye' , 'Expected IRCompressedField value to be "goodbye" after setting to "goodbye". Got: %s' %(repr(obj.value, ))

        obj.value = helloWorldBytes

        assert obj.value == helloWorldBytes , 'Expected IRCompressedField value to be b"\\x01Hello World\\x01" after setting to "\\x01Hello World\\x01"'

        
        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))



        assert dictConverted['value'] == helloWorldBytes, 'Expected asDict(forStorage=False) to contain IRCompressedField value as uncompressed bytes string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == compressedHelloWorldGzip, 'Expected asDict(forStorage=True) to contain IRCompressedField value as compressed string. \nExpected: %s\nGot:     %s' %(repr(compressedHelloWorldGzip), repr(dictForStorage['value']) )

        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating'

        assert updatedFields['value'][0] == irNull , 'Expected old value to be irNull in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == helloWorldBytes , 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving.'

        fetchObj = Model.objects.filter(name='one').first()


        assert fetchObj , 'Expected to be able to fetch object on name="one" after saving.'

        obj = fetchObj

        assert obj.value == helloWorldBytes , 'Expected value of fetched to be uncompressed value. Got: %s' %(repr(helloWorldBytes), repr(fetchObj.value), )

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after fetching'

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == helloWorldBytes, 'After fetching, Expected asDict(forStorage=False) to contain IRCompressedField value as uncompressed bytes string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == compressedHelloWorldGzip, 'After fetching, Expected asDict(forStorage=True) to contain IRCompressedField value as compressed bytes string. \nExpected: %s\nGot:     %s' %(repr(compressedHelloWorldGzip), repr(dictForStorage['value']) )

        obj.value = b'q123'

        compressedQ123 = zlib.compress(b'q123', 9)


        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating on fetched object'

        assert updatedFields['value'][0] == helloWorldBytes , 'Expected old value to be %s in updatedFields. Got: %s' %(repr(helloWorldBytes), repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == b'q123' , 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == b'q123', 'After fetching, then updating, Expected asDict(forStorage=False) to contain IRCompressedField value as uncompressed bytes string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == compressedQ123 , 'After fetching, then updating, Expected asDict(forStorage=True) to contain IRCompressedField value as compressed bytes string.\nExpected: %s\nGot:     %s' %(repr(compressedQ123), repr(dictForStorage['value']) )

        
        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving'

    def test_defaultValue(self):

        Model = self.model

        obj = Model()

        assert obj.value == b'woobley' , 'Expected defaultValue to be applied to a bytes field.\nExpected: b"woobley"\nGot:     %s' %(repr(obj.value), )

        obj.name = 'test'

        obj.save()

        assert obj.value == b'woobley' , 'Expected defaultValue to remain on a bytes field after saving'

        objFetched = Model.objects.filter(name='test').first()
        assert objFetched , 'Expected to be able to fetch object'

        obj = objFetched

        assert obj.value == b'woobley' , 'Expected defaultValue to remain on a bytes field after fetching'

        obj.value = b'cheesy'

        obj.save()

        objFetched = Model.objects.filter(name='test').first()
        assert objFetched , 'Expected to be able to fetch object'

        obj = objFetched

        assert obj.value == b'cheesy' , 'Expected to be able to change value from default.'

    def test_compressBz2(self):

        someStr = "The quick brown fox jumped over the lazy dog.\n" * 5
        someStrBytes = tobytes(someStr)

        someStrBz2 = bz2.compress(tobytes(someStr), 9)

        Model = self.model

        obj = Model()

        obj.name = 'one'

        assert obj.value == irNull , 'Expected default value (irNull) to be taken seriously.'

        obj.value = someStr

        assert obj.value == someStr

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == someStr , 'Expected original string to be retained on forStorage=False dict. Got: %s' %(repr(dictConverted['value']), )

        assert dictForStorage['value'] == someStrBz2 , 'Expected bz2 compressed value to be set on forStorage=True dict.\nExpected: %s\nGot:     %s' %(repr(someStrBz2), repr(dictForStorage['value']) )

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save object with bz2 compression'

        objFetched = Model.objects.filter(name='one').first()

        assert objFetched , 'Failed to fetch object'

        obj = objFetched

        assert obj.value == someStrBytes , 'Expected fetched object to contain the uncompressed value. Got: %s' %(repr(obj.value), )

    def test_index(self):

        Model = self.model

        obj1 = Model(name='one', value='Hello World')
        obj2 = Model(name='two', value='Goodbye World')

        ids = obj1.save()
        assert ids and ids[0] , 'Failed to save model'

        ids = obj2.save()
        assert ids and ids[0] , 'Failed to save model'


        fetchedObjs = Model.objects.filter(value='Hello World').all()

        assert fetchedObjs , 'Failed to fetch on IRCompressedField index'
        assert len(fetchedObjs) == 1 , 'Fetched wrong number of objects.'

        assert fetchedObjs[0].name == 'one' , 'Fetched wrong object'

        obj3 = Model(name='three', value='Hello World')
        
        ids = obj3.save()
        assert ids and ids[0] , 'Failed to save model'

        fetchedObjs = Model.objects.filter(value='Hello World').all()

        assert fetchedObjs , 'Failed to fetch on IRCompressedField index'
        assert len(fetchedObjs) == 2 , 'Fetched wrong number of objects.'

        assert list(sorted( [obj.name for obj in fetchedObjs] ) ) == ['one', 'three'] , 'Fetched wrong objects'

    def test_compressLzma(self):
        if not self.model:
            sys.stderr.write('NOTE: lzma compression is not available in this installation (python2 and missing an external lzma module). Cannot run test_compressLzma.\n')
            return

        
        Model = self.model

        Model.validateModel()

        lzma = Model.FIELDS['value'].getCompressMod()

        someStr = "The quick brown fox jumped over the lazy dog.\n" * 5
        someStrBytes = tobytes(someStr)

        someStrLzma = lzma.compress(tobytes(someStr))


        obj = Model()

        obj.name = 'one'

        assert obj.value == irNull , 'Expected default value (irNull) to be taken seriously.'

        obj.value = someStr

        assert obj.value == someStr

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == someStr , 'Expected original string to be retained on forStorage=False dict. Got: %s' %(repr(dictConverted['value']), )

        assert dictForStorage['value'] == someStrLzma , 'Expected lzma compressed value to be set on forStorage=True dict.\nExpected: %s\nGot:     %s' %(repr(someStrLzma), repr(dictForStorage['value']) )

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save object with lzma compression'

        objFetched = Model.objects.filter(name='one').first()

        assert objFetched , 'Failed to fetch object'

        obj = objFetched

        assert obj.value == someStrBytes , 'Expected fetched object to contain the uncompressed value. Got: %s' %(repr(obj.value), )



if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestDiffModels - Test diffing models
#

# vim: set ts=4 sw=4 st=4 expandtab

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

import base64
import sys
import subprocess

from IndexedRedis import IndexedRedisModel, irNull
from IndexedRedis.compat_str import tobytes
from IndexedRedis.fields import IRField, IRPickleField

# vim: ts=4 sw=4 expandtab

class TestDiffModels(object):
    '''
        TestDiffModels - Test IndexedRedisModel.diff
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.models = {}


        class Model_General(IndexedRedisModel):
            
            FIELDS = [
                IRField('name'),
                IRField('intVal', valueType=int),
                IRPickleField('pickledData'),
            ]

            INDEXED_FIELDS = ['name']

            KEY_NAME = 'TestDiffModels__Model_General1'

        self.models['GeneralModel'] = Model_General

        class Model_GeneralSub(Model_General):

            KEY_NAME = 'TestDiffModels__Model_GeneralSub1'

        self.models['GeneralSubModel'] = Model_GeneralSub

        class Model_SameFieldsGeneral(IndexedRedisModel):
            
            FIELDS = [
                IRField('name'),
                IRField('intVal', valueType=int),
                IRPickleField('pickledData'),
            ]

            INDEXED_FIELDS = ['name']

            KEY_NAME = 'TestDiffModels__Model_SameFieldsGeneral1'

        self.models['GeneralSameFieldsModel'] = Model_SameFieldsGeneral

        class Model_DifferentFields1(IndexedRedisModel):
            
            FIELDS = [
                IRField('name'),
                IRField('blargie', valueType=int),
                IRField('cheese', valueType=bool),
            ]

            INDEXED_FIELDS = ['name']

            KEY_NAME = 'TestDiffModels__Model_DifferentFields1'

        self.models['DifferentFieldsModel1'] = Model_DifferentFields1

        class Model_DifferentFieldProperty1(IndexedRedisModel):
            
            FIELDS = [
                IRField('name'),
                IRField('intVal', valueType=bool),
                IRPickleField('pickledData'),
            ]

            INDEXED_FIELDS = ['name']

            KEY_NAME = 'TestDiffModels__Model_DifferentFieldProperty1'

        self.models['DifferentFieldsProperty1'] = Model_DifferentFieldProperty1


        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.models:
            for model in self.models.values():
                model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.KEEP_DATA is False and self.models:
            for model in self.models.values():
                model.deleter.destroyModel()



    def test_sameModelDiff(self):
        
        GeneralModel = self.models['GeneralModel']

        pickleableData = ['one', 'two', 'three']

        obj1 = GeneralModel(name='myname', intVal=5, pickledData=pickleableData)

        obj2 = GeneralModel()
        obj2.name = 'myname'
        obj2.intVal = 5
        obj2.pickledData = pickleableData

        diffResults = obj1.diff(obj2)
        assert diffResults == {} , 'Expected two unsaved objects, same values to return blank diff result.\nGot: %s' %(repr(diffResults), )

        try:
            diffResults = IndexedRedisModel.diff(obj1, obj2)
        except Exception as e:
            raise AssertionError('Got exception trying to invoke diff via static ( IndxedRedisModel.diff(obj1, obj2) ).\n%s:   %s' %( e.__class__.__name__, str(e)) )

        assert diffResults == {} ,  'Expected two unsaved objects, same values to return blank diff result (using static invocation).\nGot: %s' %(repr(diffResults), )

        diffResults = obj1.diff(obj2, includeMeta=True)
        assert diffResults == {} , 'Expected with includeMeta=True two unsaved objects have no differences.\nGot: %s' %(repr(diffResults), )

        obj2.intVal = 12

        diffResults = obj1.diff(obj2)

        assert 'intVal' in diffResults , 'Expected after changing intVal, it would show up in diff results.'
        assert list(diffResults.keys()) == ['intVal'] , 'Expected only "intVal" to show up in keys.\nGot: %s' %(repr(list(diffResults.keys())), )
        assert diffResults['intVal'] == (5, 12) , 'Expected diffResults["intVal"] to contain (before, after).\nGot: %s' %(repr(diffResults['intVal']), )

        assert IndexedRedisModel.diff(obj1, obj2) == diffResults , 'Expected static invocation to return same results as instance call'

        ids = obj1.save()
        assert ids and ids[0] , 'Failed to save object'
        ids = obj2.save()
        assert ids and ids[0] , 'Failed to save object'

        diffResults = obj1.diff(obj2, includeMeta=True)

        diffResultsKeys = list(diffResults.keys())
        assert len(diffResultsKeys) == 2 , 'Expected to have two items in diff'
        assert '_id' in diffResultsKeys , 'Expected after save includeMeta=True to include "_id" in diffs'
        assert 'intVal' in diffResultsKeys , 'Expected after save for "intVal" to be in diffs.'

        expectedIds = ( obj1.getPk(), obj2.getPk() )
        assert diffResults['_id'] == expectedIds , 'Expected "_id" to contain first and second id.\n\nExpected: %s\n\nGot:     %s\n' %(repr(expectedIds), repr(diffResultsKeys['_id'])) 

        expectedIntVals = (5, 12)
        assert diffResults['intVal'] == expectedIntVals , 'Expected "intVal" to contain the first and second intVals.\n\nExpected: %s\n\nGot:     %s\n' %( repr(expectedIntVals), repr(diffResults['intVal']) )

        diffResults = obj2.diff(obj1)

        assert diffResults['intVal'] == (12, 5) , 'Expected comparing obj2 to obj1 switches order.'

        obj1.name = 'Becky'

        diffResults = obj1.diff(obj2)

        assert 'name' in diffResults , 'Expected after changing "name" for it to show up in diffs'

        assert diffResults['name'] == ('Becky', 'myname') , 'Expected changing "name" to show up in diffs'
        assert diffResults['intVal'] == expectedIntVals , 'Expected "intVal" to contain the first and second intVals.\n\nExpected: %s\n\nGot:     %s\n' %( repr(expectedIntVals), repr(diffResults['intVal']) )


    # TODO: Do other tests ( inherited model, different model, etc )


if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



# Copyright (c) 2014, 2015, 2016, 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields - Some types and objects related to advanced fields
#


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

__all__ = ('IRField', 'IRNullType', 'irNull', 'IRPickleField', 
	'IRCompressedField', 'IRUnicodeField', 'IRRawField', 'IRBase64Field', 
	'IRFixedPointField', 'IRDatetimeValue', 'IRJsonValue', 
	'IRBytesField', 'IRClassicField',
	'IRForeignLinkFieldBase', 'IRForeignLinkField', 'IRForeignMultiLinkField',
	'IR_NULL_STR', 'IR_NULL_BYTES', 'IR_NULL_UNICODE', 'IR_NULL_STRINGS' )

import sys
from datetime import datetime

from ..compat_str import to_unicode, tobytes
from ..deprecated import deprecatedMessage, deprecated

from .null import irNull, IR_NULL_STRINGS, IR_NULL_STR, IR_NULL_BYTES, IR_NULL_UNICODE, IRNullType

from hashlib import md5


try:
	unicode
except NameError:
	unicode = str


class IRField(str):
	'''
		IRField - An advanced field

		If a type is defined other than default/str/bytes/None , an empty value (empty string in Redis) will be assigned to the IRNullType instance provided in this module, irNull.
		irNull does not equal anything except irNull (or another IRNullType). Use this to check if a value has been assigned for other types.

		BE VERY CAREFUL ABOUT USING "float" as a type! It is an inprecise field and can vary from system to system. Instead of using a float,
		consider using fields.IRFixedPointField, which is indexable.

		An IRField may be indexable (depending on the type), and has the option to hash the index
		
	'''

	'''
	   CAN_INDEX - Set this to True if this type can be indexed. Otherwise, set it to False to disallow indexing on this field.
            The object itself is checked, so if a field is generally indexable except under certain conditions, the class can have
              True while the specific object that should be disallowed can be False.
        
         If IRField base class is used, the following types are CAN_INDEX=True: str, unicode, int, bool.  Otherwise, if CAN_INDEX is defined
           on the type, that value will be used.
	'''
	CAN_INDEX = False


	'''
		hashIndex - Whether the index field is hashed before storage / query.
			Some fields may force this value to True, others may provide the option in the constructor.
	'''
	hashIndex = False

	'''
		defaultValue - The default value used by this field.
	'''
	defaultValue = irNull


	def __init__(self, name='', valueType=str, defaultValue=irNull, hashIndex=False):
		'''
			__init__ - Create an IRField. Use this directly in the FIELDS array for advanced functionality on a field.

			@param name <str> - The name of this field

			@param valueType <type> - The type that will be used for this field. Default str/unicode (and bytes on python2)
				act the same as non-IRField FIELDS entries (just plain string), i.e. they are encoded to unicode to and from Redis.

				If you pass in None, then no decoding will take place (so whatever you provide goes in, and bytes come out of Redis).
				This is similar to IRFieldRaw

				On python3, if you pass bytes, than the field will be left as bytes.
				To be both python2 and python3 compatible, however, you can use IRBytesField

				If bool is used, then "1" and "true" are True, "0" and "false" are False, any other value is an exception.

				You can also pass an advanced type (see IndexedRedis.fields.FieldValueTypes) like datetime and json.

				All types other than string/unicode/bytes/None will be assigned 'irNull' if they were not provided a value.
				@see irNull - Equals only irNull (or other IRNullType instances). You can use this to check if an integer is defined versus 0, for example.

				While this class is create for primitive types (like int's and datetimes), more complex types extend IRField (such as pickle, compressed, or unicode with a specific encoding).

			@param defaultValue <any> (default irNull) - The value assigned to this field as a "default", i.e. when no value has yet been set. Generally, it makes sense to keep this as irNull, but you may want a different default.

			@param hashIndex <bool> (default False) - If true, the md5 hash of the value will be used for indexing and filtering. This may be useful for very long fields.

			An IRField may be indexable (depending on the type), and has the option to hash the index

			NOTE: If you are extending IRField, you should probably not call this __init__ function. So long as you implement your own "convert", any fields used are set on a class-level.
		'''
		self.defaultValue = defaultValue
		self.hashIndex = hashIndex

		if valueType in (str, unicode):
			valueType = str
			self._fromStorage = self._convertStr
			self._fromInput = self._convertStr
			self._toStorage = self._convertStr
		elif bytes != str and valueType == bytes:
			valueType = bytes
			self._fromStorage = self._convertBytes
			self._fromInput = self._convertBytes
			self._toStorage = self._convertBytes

			# Cannot index here, but CAN index if using IRBytesField. This is because python2 and python3 could handle it differently in certain cases.
			self.CAN_INDEX = False
		elif valueType in (None, type(None)):
			self._fromStorage = self._noConvert
			self._fromInput = self._noConvert
			self._toStorage = self._noConvert
			self.CAN_INDEX = False
		# I don't like these next two conditions, but it will train folks to use the correct types (whereas they may just try to shove dict in, and give up that it doesn't work)
		elif valueType in (dict, list, tuple):
#			deprecatedMessage('WARNING: Implicitly converting IRField(%s, valueType=%s) to IRField(%s, valueType=IndexedRedis.fields.FieldValueTypes.IRJsonValue)\n' %(repr(name), valueType.__name__, repr(name)), printStack=True)
			valueType = IRJsonValue
			self.CAN_INDEX = IRJsonValue.CAN_INDEX
		elif valueType == datetime:
#			deprecatedMessage('WARNING: Implicitly converting IRField(%s, valueType=datetime.datetime) to IRField(%s, valueType=IndexedRedis.fields.FieldValueTypes.IRDatetimeValue)\n' %(repr(name), repr(name)), printStack=True)
			valueType = IRDatetimeValue
			self.CAN_INDEX = IRDatetimeValue.CAN_INDEX
		else:
			if not isinstance(valueType, type):
				raise TypeError('valueType %s is not a type. Use int, str, etc' %(repr(valueType,)))
			if valueType == bool:
				self._fromStorage = self._convertBool
				self._fromInput = self._convertBool
			elif isinstance(valueType, (set, frozenset, )):
				raise TypeError('set types are not supported types. Use IRPickleField to store pickles of any type (which allow storing objects, etc), or use IRField(.. valueType=IRJsonValue) to store basic data (strings, integers) in lists.')
		self.valueType = valueType

		if valueType in (str, unicode, int, bool):
			self.CAN_INDEX = True
		elif hasattr(valueType, 'CAN_INDEX'):
			self.CAN_INDEX = valueType.CAN_INDEX
		# XXX: Commented because default CAN_INDEX is False.
#		elif valueType == float:
#			# Floats are not filterable/indexable across platforms, as they may have different rounding issues, or different number
#			#  of points of accuracy, etc.
#			# Use fields.IRFixedPointField if you need to index/filter on a floating point value.
#			self.CAN_INDEX = False
#		elif issubclass(valueType.__class__, object):
#			# Don't allow objects to index by default unless they define CAN_INDEX to be True
#			self.CAN_INDEX = False


	def toStorage(self, value):
		'''
			toStorage - Convert the value to a string representation for storage.

			  The default implementation will work here for basic types.

			@param value - The value of the item to convert
			@return A string value suitable for storing.
		'''
		if value == irNull or None:
			return IR_NULL_STR
		
		return self._toStorage(value)

	def _toStorage(self, value):
		'''
			_toStorage - Convert the value to a string for storage.

			The default implementation works for most valueTypes within IRField, override this for extending types.

			You don't need to handle null

			@param value - Value of item to convert

			@return - A string value suitable for storing
		'''
		return to_unicode(value)


	def fromStorage(self, value):
		'''
			fromStorage - Convert the value from storage to the value type.

			@param value - Value to convert

			@return - The converted value
		'''
		if value in IR_NULL_STRINGS:
			return irNull

		return self._fromStorage(value)

	def _fromStorage(self, value):
		'''
			_fromStorage - Convert the value from storage to the value type.

			  This default impl works fine for most value types, should be implemented by extending types.

			  @param value - Value to convert

			  @return - Converted value
		'''
		return self.valueType(value)

	@deprecated("convert is deprecated. Please use fromStorage instead. See fromStorage docstring if extending.")
	def convert(self, value):
		return self.fromStorage(value)


	def fromInput(self, value):
		'''
			fromInput - Convert the value from input (like assigning this through constructor or as an item assignment on the object

			@param value - Value to convert

			@return - Converted value
		'''
		if value == irNull:
			return irNull

		return self._fromInput(value)

	def _fromInput(self, value):
		'''
			_fromInput - Convert the value from input. Implement this in extending types.

			@param value - Value to convert

			@return converted value
		'''
		return self.valueType(value)


	'''
		_toIndex - Prepare a value for index.

		  Does not need to handle nulls or hashing.

		  You probably won't need to implement this, but IRFieldChain for example does.
	'''
	_toIndex = toStorage


	def toIndex(self, value):
		'''
			toIndex - An optional method which will return the value prepped for index.

			By default, "toStorage" will be called. If you provide "hashIndex=True" on the constructor,
			the field will be md5summed for indexing purposes. This is useful for large strings, etc.
		'''
		if self._isIrNull(value):
			ret = IR_NULL_STR
		else:
			ret = self._toIndex(value)

		if self.isIndexHashed is False:
			return ret

		return md5(tobytes(ret)).hexdigest()

	def getDefaultValue(self):
		'''
			getDefaultValue - Gets the default value associated with this field.

			  This is the value used when no value has been explicitly set.

			  @return - The default value
		'''
		return self.defaultValue

	@property
	def isIndexHashed(self):
		'''
			isIndexHashed - Returns if the index value should be hashed

			@return <bool> - True if this field should be hashed before indexing / filtering
		'''
		return bool(self.hashIndex)

	@property
	def name(self):
		'''
			name - Property, return this field's name

			@return <str> - Field name
		'''
		return str(self)

	def _convertStr(self, value):
		return to_unicode(value)

	def _convertBytes(self, value):
		return tobytes(value)
	

	def _noConvert(self, value):
		return value

	def _convertBool(self, value):
		if self._isNullValue(value):
			return irNull
		xvalue = value.lower()
		if xvalue in ('true', '1'):
			return True
		elif xvalue in ('false', '0'):
			return False

		# I'm not sure what to do here... Should we raise an exception because the data is invalid? Should just return True?
		raise ValueError('Unexpected value for bool type: %s' %(value,))


	@staticmethod
	def _isNullValue(value):
		'''
			_isNullValue - Tests value if it should be represented by irNull.

			convert and toStorage should test if value is null and return null (for most types)
		'''
		return bool(value in (b'', '', irNull, None) or value in IR_NULL_STRINGS )
	
	@staticmethod
	def _isIrNull(value):
		return bool( value == irNull or value in IR_NULL_STRINGS )

	def _getReprProperties(self):
		'''
			_getReprProperties - Get the properties of this field to display in repr().

				These should be in the form of $propertyName=$propertyRepr

				The default IRField implementation handles just the "hashIndex" property.

				defaultValue is part of "__repr__" impl. You should just extend this method
				with your object's properties instead of rewriting repr.

		'''
		ret = []
		if getattr(self, 'valueType', None) is not None:
			ret.append('valueType=%s' %(self.valueType.__name__, ))
		if hasattr(self, 'hashIndex'):
			ret.append('hashIndex=%s' %(self.hashIndex, ))

		return ret


	def __repr__(self):
		'''
			__repr__ - Return an object-representation string of this field instance.

			You should NOT need to extend this on your IRField, instead just implement _getReprProperties

			  to return your type's specific properties associated with this instance.

			  @see _getReprProperties
		'''
		ret = [ self.__class__.__name__, '( ', '"%s"' %(str(self), ) ]

		reprProperties = self._getReprProperties()

		defaultValue = self.getDefaultValue()
		if defaultValue == irNull:
			reprProperties.append('defaultValue=irNull')
		else:
			reprProperties.append('defaultValue=%s' %(repr(defaultValue), ))

		ret.append(', ')
		ret.append(', '.join(reprProperties))

		ret.append(' )')
		
		return ''.join(ret)

	def copy(self):
		'''
			copy - Create a copy of this IRField.

			  Each subclass should implement this, as you'll need to pass in the args to constructor.

			@return <IRField (or subclass)> - Another IRField that has all the same values as this one.
		'''
		return self.__class__(name=self.name, valueType=self.valueType, defaultValue=self.defaultValue, hashIndex=self.hashIndex)

	def __new__(self, name='', valueType=None, defaultValue=irNull, hashIndex=False):
		return str.__new__(self, name)


from .compressed import IRCompressedField
from .pickle_field import IRPickleField
from .unicode_field import IRUnicodeField
from .raw import IRRawField
from .chain import IRFieldChain
from .b64 import IRBase64Field
from .fixedpoint import IRFixedPointField
from .bytes_field import IRBytesField
from .foreign import IRForeignLinkField, IRForeignLinkFieldBase, IRForeignMultiLinkField

from .classic import IRClassicField

from .FieldValueTypes import IRDatetimeValue, IRJsonValue

# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRForeignLinkField - Test default values for fields
#

# vim: set ts=4 sw=4 st=4 expandtab

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

import base64
import sys
import subprocess

from IndexedRedis import IndexedRedisModel, irNull
from IndexedRedis.compat_str import tobytes
from IndexedRedis.fields import IRForeignLinkField, IRField, IRForeignLinkField

# vim: ts=4 sw=4 expandtab

class TestIRForeignLinkField(object):
    '''
        TestIRForeignLinkField - Test base64 field
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.models = {}

        class Model_RefedModel(IndexedRedisModel):

            FIELDS = [
                IRField('name'),
                IRField('strVal'),
                IRField('intVal', valueType=int)
            ]

            INDEXED_FIELDS = ['name']

            KEY_NAME = 'TestIRForeignLinkField__RefedModel1'

        self.models['RefedModel'] = Model_RefedModel

        class Model_MainModel(IndexedRedisModel):
            
            FIELDS = [
                IRField('name'),
                IRField('value'),
                IRForeignLinkField('other', Model_RefedModel),
            ]

            INDEXED_FIELDS = ['name']

            KEY_NAME='TestIRForeignLinkField__MainModel1'

        self.models['MainModel'] = Model_MainModel

        if testMethod == self.test_filterOnModel:
            class Model_MainModelIndexed(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRField('value'),
                    IRForeignLinkField('other', Model_RefedModel),
                ]

                INDEXED_FIELDS = ['name', 'other']

                KEY_NAME='TestIRForeignLinkField__MainModelIndexed1'

            self.models['MainModel'] = Model_MainModelIndexed

        if testMethod in (self.test_cascadeSave, self.test_cascadeFetch, self.test_reload):
            class Model_PreMainModel(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRField('value'),
                    IRForeignLinkField('main', Model_MainModel),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRForeignLinkField__PreMainModel1'

            self.models['PreMainModel'] = Model_PreMainModel


        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.models:
            for model in self.models.values():
                model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.KEEP_DATA is False and self.models:
            for model in self.models.values():
                model.deleter.destroyModel()


    def test_general(self):
        
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']

        refObj = RefedModel(name='rone', strVal='hello', intVal=1)
        ids = refObj.save(cascadeSave=False)
        assert ids and ids[0]

        mainObj = MainModel(name='one', value='cheese', other=ids[0])

        mainObj.save(cascadeSave=False)

        assert isinstance(mainObj.other, RefedModel) , 'Expected access of object to return object'

        assert mainObj.other__id == ids[0] , 'Expected __id access to return the object\'s id.'

        fetchedObj = MainModel.objects.filter(name='one').first()

        assert fetchedObj , 'Failed to fetch object'

        assert isinstance(fetchedObj.other, RefedModel) , 'After save-and-fetch, expected access of object to return object'

        assert fetchedObj.other__id == ids[0] , 'After save-and-fetch, expected __id access to return the object\'s id.'

    
    def test_disconnectAssociation(self):

        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']

        refObj = RefedModel(name='rone', strVal='hello', intVal=1)
        ids = refObj.save(cascadeSave=False)
        assert ids and ids[0]

        mainObj = MainModel(name='one', value='cheese', other=ids[0])

        mainObj.save(cascadeSave=False)

        assert isinstance(mainObj.other, RefedModel) , 'Expected access of object to return object'

        assert mainObj.other__id == ids[0] , 'Expected __id access to return the object\'s id.'

        mainObj.other = None

        assert not mainObj.other , 'After setting to None, other should be False.'

        ids = mainObj.save()
        assert ids and ids[0] , 'Failed to update'

        fetchedObj = MainModel.objects.filter(name='one').first()

        assert fetchedObj , 'Failed to fetch object'

        assert fetchedObj.other == irNull , 'Expected "other" to go back to irNull'



    def test_cascadeFetch(self):
        
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']
        PreMainModel = self.models['PreMainModel']

        refObj = RefedModel(name='rone', strVal='hello', intVal=1)
        ids = refObj.save(cascadeSave=False)
        assert ids and ids[0]

        mainObj = MainModel(name='one', value='cheese', other=ids[0])

        ids = mainObj.save(cascadeSave=False)
        assert ids and ids[0] , 'Failed to save object'

        preMainObj = PreMainModel(name='pone', value='bologna')
        preMainObj.main = mainObj

        ids = preMainObj.save(cascadeSave=False)
        assert ids and ids[0], 'Failed to save object'

        objs = PreMainModel.objects.filter(name='pone').all(cascadeFetch=True)

        assert objs and len(objs) == 1 , 'Failed to fetch single PreMainModel object'

        obj = objs[0]

        oga = object.__getattribute__

        assert oga(obj, 'main').isFetched() is True , 'Expected cascadeFetch to fetch sub object. Failed one level down (not marked isFetched)'

        assert oga(obj, 'main').obj , 'Expected cascadeFetch to fetch sub object. Failed one level down (object not present)'

        fetchedMainObj = oga(obj, 'main').obj

        assert fetchedMainObj == mainObj , 'Fetched MainModel object has wrong values.\n\nExpected: %s\n\nGot:     %s\n' %(repr(mainObj), repr(fetchedMainObj))

        assert oga(fetchedMainObj, 'other').isFetched() is True , 'Expected cascadeFetch to fetch sub object. Failed two levels down (not marked isFetched)'
        assert oga(fetchedMainObj, 'other').obj , 'Expected cascadeFetch to fetch sub object. Failed to levels down (object not present)'

        assert oga(fetchedMainObj, 'other').obj.name == 'rone' , 'Missing values on two-level-down fetched object.'

        MainModel.deleter.destroyModel()
        RefedModel.deleter.destroyModel()
        PreMainModel.deleter.destroyModel()

        # Now test that cascadeFetch=False does NOT fetch the subs.
        refObj = RefedModel(name='rone', strVal='hello', intVal=1)
        ids = refObj.save(cascadeSave=False)
        assert ids and ids[0]

        mainObj = MainModel(name='one', value='cheese', other=ids[0])

        ids = mainObj.save(cascadeSave=False)
        assert ids and ids[0] , 'Failed to save object'

        preMainObj = PreMainModel(name='pone', value='bologna')
        preMainObj.main = mainObj

        ids = preMainObj.save(cascadeSave=False)
        assert ids and ids[0], 'Failed to save object'

        objs = PreMainModel.objects.filter(name='pone').all(cascadeFetch=False)

        assert objs and len(objs) == 1 , 'Failed to fetch objects'

        obj = objs[0]

        assert oga(obj, 'main').isFetched() is False , 'Expected cascadeFetch=False to NOT automatically fetch sub object. isFetched is marked True one level down.'
        assert not bool(oga(obj, 'main').obj) , 'Expected cascadeFetch=False to NOT automatically fetch sub object. Object was found on sub object one level down.'


        # Now try the IndexedRedisModel.cascadeFetch function to perform the fetch.
        objs = PreMainModel.objects.filter(name='pone').all(cascadeFetch=False)

        assert objs and len(objs) == 1 , 'Failed to fetch objects'

        obj = objs[0]
        obj.cascadeFetch()

        assert oga(obj, 'main').isFetched() is True , 'Expected Model.cascadeFetch to fetch sub object. Failed one level down (not marked isFetched)'

        assert oga(obj, 'main').obj , 'Expected Model.cascadeFetch to fetch sub object. Failed one level down (object not present)'

        mainObj = oga(obj, 'main').obj

        assert oga(mainObj, 'other').isFetched() is True , 'Expected Model.cascadeFetch to fetch sub object. Failed two levels down (not marked isFetched)'
        assert oga(mainObj, 'other').obj , 'Expected Model.cascadeFetch to fetch sub object. Failed to levels down (object not present)'

        assert oga(mainObj, 'other').obj.name == 'rone' , 'Missing values on two-level-down fetched object.'




    def test_assign(self):
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']

        refObj1 = RefedModel(name='rone', strVal='hello', intVal=1)
        refObj2 = RefedModel(name='rtwo', strVal='world', intVal=2)

        ids1 = refObj1.save(cascadeSave=False)
        assert ids1 and ids1[0] , 'Failed to save object'

        ids2 = refObj2.save(cascadeSave=False)
        assert ids2 and ids2[0] , 'Failed to save object'

        mainObj = MainModel(name='one', value='cheese', other=ids1[0])

        assert mainObj.other.hasSameValues(refObj1) , 'Expected other with id of refObj1 to link to refObj1'

        mainObj.other = ids2[0]

        assert mainObj.other.hasSameValues(refObj2) , 'Expected other with id of refObj2 to link to refObj2'

        ids = mainObj.save(cascadeSave=False)
        assert ids and ids[0] , 'Failed to save object'

        fetchedObj = mainObj.objects.filter(name='one').first()

        assert fetchedObj , 'Failed to fetch object'

        assert fetchedObj.other__id == ids2[0] , 'Expected __id access to return the object\'s id.'


        mainObj = fetchedObj

        firstRefObj = RefedModel.objects.filter(name='rone').first()

        assert firstRefObj , 'Failed to fetch object'

        mainObj.other = firstRefObj

        ids = mainObj.save(cascadeSave=False)
        assert ids and ids[0] , 'Failed to save'

        fetchedObj = mainObj.objects.filter(name='one').first()

        assert fetchedObj , 'Failed to fetch object'

        assert fetchedObj.other__id == ids1[0] , 'Expected save using Model object would work properly. Did not fetch correct id after save.'
        
    def test_filterOnModel(self):
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']

        refObj1 = RefedModel(name='rone', strVal='hello', intVal=1)
        refObj2 = RefedModel(name='rtwo', strVal='world', intVal=2)

        ids1 = refObj1.save(cascadeSave=False)
        assert ids1 and ids1[0] , 'Failed to save object'

        ids2 = refObj2.save(cascadeSave=False)
        assert ids2 and ids2[0] , 'Failed to save object'

        mainObj = MainModel(name='one', value='cheese', other=ids1[0])

        assert object.__getattribute__(mainObj, 'other').isFetched() is False , 'Expected object not to be fetched before access'

        assert mainObj.other.hasSameValues(refObj1) , 'Expected other with id of refObj1 to link to refObj1'

        assert object.__getattribute__(mainObj, 'other').isFetched() is True, 'Expected object to be fetched after access'
        mainObj.other = ids2[0]

        assert mainObj.other.hasSameValues(refObj2) , 'Expected other with id of refObj2 to link to refObj2'

        ids = mainObj.save(cascadeSave=False)

        fetchedObjs = MainModel.objects.filter(other=ids2[0]).all()

        assert fetchedObjs and len(fetchedObjs) == 1 , 'Expected to be able to filter on numeric pk'

        fetchedObjs = MainModel.objects.filter(other=refObj2).all()

        assert fetchedObjs and len(fetchedObjs) == 1 , 'Expected to be able to filter on object itself'



    def test_cascadeSave(self):
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']
        PreMainModel = self.models['PreMainModel']

        refObj1 = RefedModel(name='rone', strVal='hello', intVal=1)

        mainObj = MainModel(name='one', value='cheese')

        mainObj.other = refObj1

        ids = mainObj.save(cascadeSave=True)

        assert ids and ids[0] , 'Failed to save object'

        assert mainObj.other._id , 'Failed to set id on other object'

        obj = MainModel.objects.filter(name='one').first()

        assert obj , 'Failed to fetch object by name'

        assert obj.other , 'Did not cascade save second object and link to parent'

        assert obj.other.name == 'rone' , 'Did save values on cascaded object'

        RefedModel.deleter.destroyModel()
        MainModel.deleter.destroyModel()

        refObj1 = RefedModel(name='rone', strVal='hello', intVal=1)
        mainObj = MainModel(name='one', value='cheese')

        mainObj.other = refObj1

        preMainObj = PreMainModel(name='pone', value='bologna')

        preMainObj.main = mainObj

        ids = preMainObj.save(cascadeSave=True)

        assert ids and ids[0] , 'Failed to save object'


        obj = PreMainModel.objects.filter(name='pone').first()

        assert obj , 'Failed to fetch object by name'

        assert obj.main , 'Failed to link one level down'
        assert obj.main.name == 'one' , 'Did not save values one level down'

        assert obj.main.other , 'Failed to link two levels down'
        assert obj.main.other.name == 'rone' , 'Failed to save values two levels down'


    def test_reload(self):
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']
        PreMainModel = self.models['PreMainModel']

        oga = object.__getattribute__

        refObj1 = RefedModel(name='rone', strVal='hello', intVal=1)
        refObj2 = RefedModel(name='rtwo', strVal='hello', intVal=2)

        refObj2.save()

        mainObj = MainModel(name='one', value='cheese')

        mainObj.other = refObj1

        ids = mainObj.save(cascadeSave=True)

        assert ids and ids[0] , 'Failed to save object'

#        mainObj = MainModel.objects.first()

        robj = RefedModel.objects.filter(name='rone').first()

        robj.intVal = 5
        ids = robj.save()
        assert ids and ids[0] , 'Failed to save object'

        mainObj.reload(cascadeObjects=False)

        assert mainObj.other.intVal != 5 , 'Expected reload(cascadeObjects=False) to NOT reload sub-object'

        mainObj.reload(cascadeObjects=True)
        
        assert mainObj.other.intVal == 5 , 'Expected reload(cascadeObjects=True) to reload sub-object'

        mainObj = MainModel.objects.first()

        assert oga(mainObj, 'other').isFetched() is False , 'Expected sub-obj to not be fetched automatically'

        z = mainObj.asDict(forStorage=False)

        assert oga(mainObj, 'other').isFetched() is False , 'Expected calling "asDict" to not fetch foreign object'

        reloadedData = mainObj.reload(cascadeObjects=False)

        assert oga(mainObj, 'other').isFetched() is False , 'Expected reload(cascadeObjects=False) to not fetch sub-object'

        assert not reloadedData , 'Expected no data to be reloaded'

        reloadedData = mainObj.reload(cascadeObjects=True)

        assert oga(mainObj, 'other').isFetched() is False , 'Expected reload(cascadeObjects=True) to not fetch sub-object'

        assert not reloadedData , 'Expected no data to be reloaded'

        mainObj.other

        assert oga(mainObj, 'other').isFetched() is True , 'Expected access to fetch sub object'

        reloadedData = mainObj.reload(cascadeObjects=False)

        assert not reloadedData , 'Expected no data to be reloaded with local resolved but unchanged.'

        mainObj.other.intVal = 99

        reloadedData = mainObj.reload(cascadeObjects=False)

        assert not reloadedData , 'Expected to not see "other" in reloaded data, as pk did not change but values did.'

        reloadedData = mainObj.reload(cascadeObjects=True)

        assert 'other' in reloadedData , 'Expected "other" to  be reloaded with reload(cascadeObjects=True)'

        assert reloadedData['other'][0].getObj().intVal == 99 , 'Expected old value to be present in reload'

        assert reloadedData['other'][1].getObj().intVal == 5 , 'Expected new value to be present in reload'


        mainObj = MainModel.objects.first()

        mainObj.other = refObj2._id

        reloadedData = mainObj.reload(cascadeObjects=False)
        
        assert 'other' in reloadedData , 'Expected to see "other" in reloaded data when cascadeObjects=False when pk changes. Using pk assignment.'

        mainObj = MainModel.objects.first()

        mainObj.other = refObj2._id
        reloadedData = mainObj.reload(cascadeObjects=True)
        
        assert 'other' in reloadedData , 'Expected to see "other" in reloaded data when cascadeObjects=True when pk changes. Using pk assignment.'


        mainObj = MainModel.objects.first()

        mainObj.other = refObj2

        reloadedData = mainObj.reload(cascadeObjects=False)
        assert 'other' in reloadedData , 'Expected to see "other" in reloaded data when cascadeObjects=False when pk changes. Using obj assignment.'

        mainObj = MainModel.objects.first()

        mainObj.other = refObj2

        reloadedData = mainObj.reload(cascadeObjects=True)
        assert 'other' in reloadedData , 'Expected to see "other" in reloaded data when cascadeObjects=True when pk changes. Using obj assignment.'

        preMainObj = PreMainModel(name='pone', value='zzz')

        preMainObj.main = mainObj._id

        preMainObj.save()


        preMainObj = PreMainModel.objects.filter(name='pone').first()

        assert preMainObj , 'Failed to fetch object'

        preMainObj.main.other.intVal = 33
        reloadedData = preMainObj.reload(cascadeObjects=False)
        assert not reloadedData , 'Expected to not see any reloaded data for cascadeObjects=False when object two-levels-down has changed values.'

        reloadedData = preMainObj.reload(cascadeObjects=True)
        assert 'main' in reloadedData , 'Expected to see "main" (one-level-down) object show up in reloaded data for cascadeObjects=True when object two-levels-down has changed values.'



    def test_suppressFetching(self):
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']

        oga = object.__getattribute__

        refObj1 = RefedModel(name='rone', strVal='hello', intVal=1)

        mainObj = MainModel(name='one', value='cheese')

        mainObj.other = refObj1

        ids = mainObj.save(cascadeSave=True)

        assert ids and ids[0] , 'Failed to save object'

        mainObj = MainModel.objects.first()

        assert oga(mainObj, 'other').isFetched() is False , 'Expected other to not be fetched right away'

        updatedFields = mainObj.getUpdatedFields()

        assert not updatedFields , 'Expected updatedFields to be blank. Got: %s' %(repr(updatedFields), )

        assert oga(mainObj, 'other').isFetched() is False , 'Expected other to not be fetched after calling getUpdatedFields'


    def test_unsavedChanges(self):
        MainModel = self.models['MainModel']
        RefedModel = self.models['RefedModel']

        oga = object.__getattribute__

        refObj1 = RefedModel(name='rone', strVal='hello', intVal=1)

        mainObj = MainModel(name='one', value='cheese')

        mainObj.other = refObj1

        ids = mainObj.save(cascadeSave=True)

        mainObj2 = MainModel(name='one', value='cheese')

        assert mainObj.hasSameValues(mainObj2) is False , 'Expected not to have same values when one has foreign set, other does not.'


        mainObj2.other = refObj1._id

        assert mainObj.hasSameValues(mainObj2) , 'Expected to have same values when one has object, other has id'

        mainObj2 = MainModel(name='one', value='cheese')
        mainObj2.other = refObj1

        assert mainObj.hasSameValues(mainObj2) , 'Expected to have same values with same object on both'

        mainObj = MainModel.objects.first()

        assert mainObj.hasSameValues(mainObj2) , 'Expected to have same values after fetch. one has id, one has object.'

        mainObj.other.intVal = 55

        assert not mainObj.hasSameValues(mainObj2) , 'Expected changing a foreign link field\'s data on one object would cause hasSameValues to be False.'

        assert mainObj.hasSameValues(mainObj2, cascadeObject=False) , 'Expected changing a foreign link field\'s data on one object would leave hasSameValues(... , cascadeObject=False) to be True'


if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



# Copyright (c) 2016, 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# IRQueryableList - QueryableList with some added callbacks to IndexedRedis
#


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

import pprint

from QueryableList import QueryableListObjs

__all__ = ('IRQueryableList', )


class IRQueryableList(QueryableListObjs):
	'''
		IRQueryableList - A QueryableList for IndexedRedis models.

		Only supports models of one type. Preferably, you should pass "mdl" to init, otherwise
		  it will infer from the first object.

		For performance, there is NOT type checking that all models provided are of the same type,
		this is your responsibility. 

		IndexedRedis will only ever return an IRQueryableList that follow these constraints (always have model defined,
		  and only one type of model ever contained), but if you explicitly append objects you need to make sure they
		  are of the correct type.
	'''

	def __init__(self, val=None, mdl=None):
		'''
			__init__ - Create this object

			@param val - None for empty list, IndexedRedisModel for a one-item list, or a list/tuple or subclass of initial values.
			@param mdl - The IndexedRedisModel that this list will contain. Provide this now if you can, otherwise it will be inferred from
			  the first item added or present in the list.

			@raises ValueError if "mdl" is not an IndexedRedisModel
		'''
		if val is None:
			QueryableListObjs.__init__(self)
		else:
			QueryableListObjs.__init__(self, val)


		self.mdl = mdl

		if not mdl:
			# If not explicitly defined, try to infer model if objects were provided.
			#  otherwise, inference will be attempted when an operation that requires it is performed.
			self.mdl = self.getModel()
		else:
			# This is called in getModel() if we did infer, so no need to call twice.
			self.__validate_model(mdl)


	@staticmethod
	def __validate_model(mdl):
		'''
			__validate_model - Internal function to check that model is of correct type.

			Uses a class variable that has been defined for IndexedRedisModel s for a long time, not the type itself, to prevent circular imports etc.
	
			@param mdl - type to validate
		'''
		if not hasattr(mdl, '_is_ir_model'):
			raise ValueError('Model %s is not an IndexedRedisModel' %(str(mdl.__class__.__name__),))

	def getModel(self):
		'''
			getModel - get the IndexedRedisModel associated with this list. If one was not provided in constructor,
			  it will be inferred from the first item in the list (if present)

			  @return <None/IndexedRedisModel> - None if none could be found, otherwise the IndexedRedisModel type of the items in this list.

			@raises ValueError if first item is not the expected type.
		'''
		if not self.mdl and len(self) > 0:
			mdl = self[0].__class__
			self.__validate_model(mdl)

			self.mdl = mdl

		return self.mdl


	def delete(self):
		'''
			delete - Delete all objects in this list.

			@return <int> - Number of objects deleted
		'''
		if len(self) == 0:
			return 0
		mdl = self.getModel()
		return mdl.deleter.deleteMultiple(self)


	def save(self):
		'''
			save - Save all objects in this list
		'''
		if len(self) == 0:
			return []
		mdl = self.getModel()
		return mdl.saver.save(self)


	def reload(self):
		'''
			reload - Reload all objects in this list. 
				Updates in-place. To just fetch all these objects again, use "refetch"

			@return - List (same order as current objects) of either exception (KeyError) if operation failed,
			  or a dict of fields changed -> (old, new)
		'''
		if len(self) == 0:
			return []

		ret = []
		for obj in self:
			res = None
			try:
				res = obj.reload()
			except Exception as e:
				res = e

			ret.append(res)

		return ret


	def refetch(self):
		'''
			refetch - Fetch a fresh copy of all items in this list.
				Returns a new list. To update in-place, use "reload".

			@return IRQueryableList<IndexedRedisModel> - List of fetched items
		'''

		if len(self) == 0:
			return IRQueryableList()

		mdl = self.getModel()
		pks = [item._id for item in self if item._id]

		return mdl.objects.getMultiple(pks)
	
	def pprint(self, stream=None):
		dicts = [obj.asDict(includeMeta=True, forStorage=False, strKeys=True) for obj in self]

		pprint.pprint(dicts, stream=stream)


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRFieldChain - Test field chaining
#

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

# vim: set ts=4 sw=4 st=4 expandtab

import base64
import sys
import subprocess

from IndexedRedis import IndexedRedisModel, irNull, InvalidModelException
from IndexedRedis.compat_str import tobytes, to_unicode, getDefaultIREncoding, setDefaultIREncoding
from IndexedRedis.fields import IRBase64Field, IRField, IRUnicodeField, IRPickleField, IRCompressedField, IRFieldChain

# vim: ts=4 sw=4 expandtab
class TestObj(object):

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __hash__(self):
        return hash( str(hash(self.a)) + '_' + str(hash(self.b)) )

    def __eq__(self, other):
        if type(other) != type(self):
            return False

        return bool(self.a == other.a and self.b == other.b)

    def __ne__(self, other):
        if type(other) != type(self):
            return True

        return bool(self.a != other.a or self.b != other.b)

    def __repr__(self):
        return 'TestObj( a=%s , b=%s )' %(repr(self.a), repr(self.b) )

    __str__ = __repr__


class TestIRFieldChain(object):
    '''
        TestIRFieldChain - Test field chaining
    '''

    KEEP_DATA = False

    def setup_class(self):
        self.defaultIREncoding = getDefaultIREncoding()

        self.utf16DataBytes = b'\xff\xfe\x01\xd8\x0f\xdc\x01\xd8-\xdc\x01\xd8;\xdc\x01\xd8+\xdc'
        self.utf16Data = self.utf16DataBytes.decode('utf-16')

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        if testMethod == self.test_base64AndUnicode:
            setDefaultIREncoding('ascii') # Ensure we are using the "encoding" value provided in the field
            class Model_Base64Unicode(IndexedRedisModel):
                
                FIELDS = [
                    IRField('name'),
                    IRFieldChain('value', [IRUnicodeField(encoding='utf-8'), IRBase64Field(encoding='utf-8')], defaultValue=irNull),
                    IRFieldChain('value2', [IRUnicodeField(encoding='utf-8'), IRBase64Field(encoding='utf-8')], defaultValue='qqz'),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRFieldChain__ModelBase64Unicode'

            self.model = Model_Base64Unicode
        elif testMethod == self.test_utf16Compression:
            setDefaultIREncoding('ascii') # Ensure we are using the "encoding" value provided in the field

            class Model_Utf16Compression(IndexedRedisModel):

                FIELDS = [
                    IRField('name'),
                    IRFieldChain('value', [IRUnicodeField(encoding='utf-16'), IRCompressedField()]),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRFieldChain__Utf16Compression'

            self.model = Model_Utf16Compression
            
        elif testMethod == self.test_compressPickle:
            class Model_CompressPickle(IndexedRedisModel):

                FIELDS = [
                    IRField('name'),
                    IRFieldChain('value', [IRPickleField(), IRCompressedField()]),
                    IRFieldChain('value2', [IRPickleField(), IRCompressedField()], defaultValue=['a', 'b', 'c'])
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRFieldChain__ModelCompressPickle'

            self.model = Model_CompressPickle
        elif testMethod == self.test_intBase64:
            class Model_IntBase64(IndexedRedisModel):
                
                FIELDS = [
                    IRField('name'),
                    IRFieldChain('value', [IRField(valueType=int), IRBase64Field ] ),
                    IRFieldChain('value2', [IRField(valueType=int), IRBase64Field ], defaultValue=-1 )
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRFieldChain__IntBase64'

            self.model = Model_IntBase64
        elif testMethod == self.test_index:
            class Model_ChainIndex(IndexedRedisModel):

                FIELDS = [
                    IRField('name'),
                    IRFieldChain('value', [IRField(valueType=str), IRBase64Field()]),
                ]

                INDEXED_FIELDS = ['name', 'value']

                KEY_NAME = 'TestIRFieldChain__ModelChainIndex'

            self.model = Model_ChainIndex

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''
        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()

        # Reset encoding back to original, some tests may change it.
        setDefaultIREncoding(self.defaultIREncoding)


    def test_base64AndUnicode(self):
        Model = self.model

        obj = Model()
    
        prettyPictures = b' \xe2\x9c\x8f \xe2\x9c\x90 \xe2\x9c\x91 \xe2\x9c\x92 \xe2\x9c\x93 \xe2\x9c\x94 \xe2\x9c\x95 \xe2\x9c\x96 \xe2\x9c\x97 \xe2\x9c\x98 \xe2\x9c\x99 \xe2\x9c\x9a \xe2\x9c\x9b \xe2\x9c\x9c \xe2\x9c\x9d \xe2\x9c\x9e \xe2\x9c\x9f \xe2\x9c\xa0 \xe2\x9c\xa1 \xe2\x9c\xa2 \xe2\x9c\xa3 \xe2\x9c\xa4 \xe2\x9c\xa5 \xe2\x9c\xa6 \xe2\x9c\xa7 \xe2\x9c\xa9 \xe2\x9c\xaa \xe2\x9c\xab '
        prettyPicturesUnicode = to_unicode(prettyPictures, encoding='utf-8')

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected no updated fields when object is first created.\nExpected: %s\nGot:     %s' %(repr([]), repr(updatedFields) )

        assert obj.value == irNull , 'Expected default value of IRFieldChain to be irNull when defaultValue=irNull. Got: %s' %(obj.value, )
        assert obj.value2 == 'qqz' , 'Expected default value of IRFieldChain to be "qqz" when defaultValue="qqz". Got: %s' %(obj.value2, )

        obj.name = 'one'

        obj.save()


        updatedFields = obj.getUpdatedFields()
        assert updatedFields == {} , 'Expected no updated fields after object is saved. Got: %s' %(repr(updatedFields), )


        obj.value = prettyPicturesUnicode
        assert obj.value == prettyPicturesUnicode , 'Expected IRFieldChain value to unicode of prettyPictures when set to unicode value'

        obj.value = prettyPictures
        assert obj.value == prettyPicturesUnicode , 'Expected IRFieldChain value to unicode of prettyPictures when set to bytes value'

        b64Value = base64.b64encode(tobytes(obj.value, encoding='utf-8'))

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to a dict, both not-for-storage and for-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == prettyPicturesUnicode, 'Expected asDict(forStorage=False) to contain IRFieldChain value as unicode str.. Got: %s' %(repr(dictConverted['value']), )
        assert dictForStorage['value'] == b64Value , 'Expected asDict(forStorage=True) to contain IRFieldChain that was base64 encoded.\nExpected: %s\nGot:     %s' %(repr(b64Value), repr(dictForStorage['value']) )

        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating'

        assert updatedFields['value'][0] == irNull , 'Expected old value to be irNull in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == prettyPicturesUnicode, 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving.'

        fetchObj = Model.objects.filter(name='one').first()


        assert fetchObj , 'Expected to be able to fetch object on name="one" after saving.'

        obj = fetchObj

        assert obj.value == prettyPicturesUnicode , 'Expected value of fetched to be unicode string. Got: <%s> %s' %(type(fetchObj.value).__name__, repr(fetchObj.value), )

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after fetching'

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to a dict, both not-for-storage and for-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == prettyPicturesUnicode, 'After fetching, Expected asDict(forStorage=False) to contain IRFieldChain value as unicode string.  Got: %s' %(repr(dictConverted['value']), )
        assert dictForStorage['value'] == b64Value , 'After fetching, Expected asDict(forStorage=True) to contain IRFieldChain that was base64 encoded.\nExpected: %s\nGot:     %s' %(repr(b64Value), repr(dictForStorage['value']) )

        obj.value = b'q123'


        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating on fetched object'

        assert updatedFields['value'][0] == prettyPicturesUnicode , 'Expected old value to be prettyPictures as unicode in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == u'q123' , 'Expected converted value to be new value in updatedFields as unicode. Got: %s' %(repr(updatedFields['value'][1]), )

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to a dict, both not-for-storage and for-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        b64Value = base64.b64encode(tobytes(obj.value))

        assert dictConverted['value'] == u'q123', 'After fetching, then updating, Expected asDict(forStorage=False) to contain IRFieldChain value as unicode string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == b64Value , 'After fetching, then updating, Expected asDict(forStorage=True) to contain IRFieldChain that was base64 encoded.\nExpected: %s\nGot:     %s' %(repr(b64Value), repr(dictForStorage['value']) )

        
        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving'

    def test_compressPickle(self):
        Model = self.model

        obj = Model()


        testObj1 = TestObj(19, 'cheese')
        testObj2 = TestObj(-4.77, 99)

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected no updated fields when object is first created.\nExpected: %s\nGot:     %s' %(repr([]), repr(updatedFields) )

        assert obj.value == irNull , 'Expected default value of IRFieldChain to be irNull when defaultValue=irNull. Got: %s' %(obj.value, )
        assert obj.value2 == ['a', 'b', 'c'] , 'Expected default value of IRFieldChain to be ["a", "b", "c"] when defaultValue= ["a", "b", "c"]. Got: %s' %(obj.value2, )

        obj.name = 'one'

        obj.save()


        updatedFields = obj.getUpdatedFields()
        assert updatedFields == {} , 'Expected no updated fields after object is saved. Got: %s' %(repr(updatedFields), )


        obj.value = testObj1
        assert obj.value == testObj1, 'Expected IRFieldChain value to retain after setting'

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to a dict, both not-for-storage and for-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == testObj1, 'Expected asDict(forStorage=False) to contain IRFieldChain value as testObj1 Got: %s' %(repr(dictConverted['value']), )
        # Note, just testing that it changed.. if we can fetch and convert it assume it was converted correctly.
        assert dictForStorage['value'] != testObj1, 'Expected asDict(forStorage=True) to be converted in some way.'

        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating'

        assert updatedFields['value'][0] == irNull , 'Expected old value to be irNull in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == testObj1, 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving.'

        fetchObj = Model.objects.filter(name='one').first()


        assert fetchObj , 'Expected to be able to fetch object on name="one" after saving.'

        obj = fetchObj

        assert obj.value == testObj1, 'Expected value of fetched to be testObj1. Got: <%s> %s' %(type(fetchObj.value).__name__, repr(fetchObj.value), )

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after fetching'

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to a dict, both not-for-storage and for-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == testObj1, 'After fetching, Expected asDict(forStorage=False) to contain IRFieldChain value as object.  Got: %s' %(repr(dictConverted['value']), )
        assert dictForStorage['value'] != testObj1 , 'After fetching, Expected asDict(forStorage=True) to contain IRFieldChain that was encoded in some way.'

        obj.value = testObj2


        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating on fetched object'

        assert updatedFields['value'][0] == testObj1, 'Expected old value to be testObj1 in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == testObj2 , 'Expected new value to be testObj2 in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to a dict, both not-for-storage and for-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == testObj2, 'After fetching, then updating, Expected asDict(forStorage=False) to contain IRFieldChain value as testObj2. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] not in (testObj1, testObj2), 'After fetching, then updating, Expected asDict(forStorage=True) to contain a converted value.'

        
        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving'

    def test_intBase64(self):
        
        Model = self.model

        obj = Model()

        assert obj.value == irNull , 'Expected default value to be irNull on chain when defaultValue=irNull. Got: <%s> %s' %(obj.value.__class__.__name__, repr(obj.value), )
        assert obj.value2 == -1 , 'Expected default value to be -1 on chain when defaultValue=-1. Got: <%s> %s' %(obj.value.__class__.__name__, repr(obj.value), )

        obj.name = 'one'

        obj.save()

        objFetched = Model.objects.filter(name='one').first()

        assert objFetched , 'Expected to be able to fetch on indexed field'

        obj = objFetched

        assert obj.value == irNull , 'After fetch-and-save, expected default value to be irNull on chain when defaultValue=irNull. Got: <%s> %s' %(obj.value.__class__.__name__, repr(obj.value), )
        assert obj.value2 == -1 , 'After fetch-and-save, expected default value to be -1 on chain when defaultValue=-1. Got: <%s> %s' %(obj.value.__class__.__name__, repr(obj.value), )


        obj.value = '14'

        assert obj.value == 14 , 'Expected value to be conveted to int after setting'

        forStorageDict = obj.asDict(forStorage=True)


        b64Value = base64.b64encode(b'14')

        assert forStorageDict['value'] == b64Value , 'Expected value to be base64 encoded in forStorage representation'

        obj.save()

        objFetched = Model.objects.filter(name='one').first()

        assert objFetched , 'Expected to be able to fetch on indexed field'

        obj = objFetched

        assert obj.value == 14 , 'Expected value to be set to 14 after fetching. Got: <%s> %s' %(obj.value.__class__.__name__, repr(obj.value))

    def test_utf16Compression(self):

        Model = self.model

        obj = Model()

        assert obj.value == irNull , 'Expected default value to be retained'

        obj.name = 'one'

        obj.save()

        obj.value = self.utf16DataBytes

        assert obj.value == self.utf16Data , 'Expected bytes data to be converted to unicode after setting on object'

        ids = obj.save()

        assert ids and ids[0] , 'Failed to save object'

        fetchedObjs = Model.objects.all()

        assert len(fetchedObjs) == 1 , 'Expected there to be one object saved, got: %d' %(len(fetchedObjs), )

        obj = fetchedObjs[0]

        assert obj.value == self.utf16Data , 'Expected data to be the utf-16 string after fetching'


    def test_index(self):

        Model = self.model

        obj1 = Model(name='one')

        obj1.value = 'Hello World'

        obj2 = Model(name='two')

        obj2.value = 'Goodbye World'

        ids = obj1.save()
        assert ids and ids[0] , 'Failed to save object'

        ids = obj2.save()
        assert ids and ids[0] , 'Failed to save object'

        objFetched = Model.objects.filter(value='Hello World').all()

        assert len(objFetched) == 1 , 'Failed to fetch one object. Got: %d' %(int(objFetched), )

        assert objFetched[0].name == 'one' , 'Fetched wrong object'

        # TODO: Clean up and test more

        class BadModel1(IndexedRedisModel):
            
            FIELDS = [
                IRField('name'),
                IRFieldChain('value', [IRPickleField(), IRCompressedField()])
            ]

            INDEXED_FIELDS = ['name', 'value']

            KEY_NAME = 'TestIRFieldChain__BadModel1'

        gotInvalidModelException = False
        try:
            BadModel1.validateModel()
        except InvalidModelException:
            gotInvalidModelException = True

        assert gotInvalidModelException , 'Expected to get an invalid model exception trying to index on an IRFieldChain which contains a non-indexable field'


        class AutoHashIndexModel1(IndexedRedisModel):
            FIELDS = [
                IRField('name'),
                IRFieldChain('value', [IRField(), IRBase64Field()] ),
            ]

            INDEXED_FIELDS = ['name', 'value']

            KEY_NAME = 'TestIRFieldChain__AutoHashIndexModel1'

        gotInvalidModelException = False
        ie = None
        try:
            AutoHashIndexModel1.validateModel()
        except InvalidModelException as _ie:
            gotInvalidModelException = True
            ie = _ie

        assert not gotInvalidModelException , 'Expected to be able to index on a field chain containing two indexable IRField types. ' + str(ie)

        assert AutoHashIndexModel1.FIELDS['value'].hashIndex is True , 'Expected hashIndex to be automatically set to True when an IRCompressedField is right-most, as it forces hashIndex=True'


if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab
